import numpy as np
from sklearn.decomposition import PCA


def _signed_angle(v1: np.ndarray, v2: np.ndarray, vn: np.ndarray) -> float:
    """
    Calculate the signed angle between two vectors in 3D space.

    Parameters:
    - v1: numpy array, first vector
    - v2: numpy array, second vector
    - vn: numpy array, normal vector

    Returns:
    - angle: float, signed angle between the two vectors
    """

    angle = np.arctan2(np.dot(vn, np.cross(v1, v2)), np.dot(v1, v2))

    return angle


def _compute_rotation_angle_and_indices(
    mask_for_pca: np.ndarray,
    target_axis: str,
    rotation_plane: str,
) -> tuple[float, tuple[int, int]]:
    """
    Compute the rotation angle and indices for axis alignment.

    Parameters:
    - mask_for_pca: numpy array, the mask array.
    - target_axis: str, the target axis for alignment ('X', 'Y', or 'Z').
    - rotation_plane: str, the rotation plane for alignment ('XY', 'XZ', or 'YZ').
    - temporal_slice: slice, the temporal slice to use for alignment.

    Returns:
    - rotation_angle: float, the rotation angle in degrees.
    - rotation_plane_indices: tuple, the indices of the rotation plane.

    """

    # Perform PCA on the mask to determine the major axis
    pca = PCA(n_components=3)
    pca.fit(np.argwhere(mask_for_pca))
    major_axis_vector = pca.components_[0]

    plane_normal_vector = {
        "XY": np.array([1, 0, 0]),
        "XZ": np.array([0, 1, 0]),
        "YZ": np.array([0, 0, 1]),
    }[rotation_plane]

    # Remove the component of the major axis vector that is perpendicular to the rotation plane
    major_axis_vector -= (
        np.dot(major_axis_vector, plane_normal_vector) * plane_normal_vector
    )
    major_axis_vector = major_axis_vector / np.linalg.norm(major_axis_vector)
    major_axis_vector *= np.sign(np.max(major_axis_vector))

    target_axis_vector = {
        "Z": np.array([1, 0, 0]),
        "Y": np.array([0, 1, 0]),
        "X": np.array([0, 0, 1]),
    }[target_axis]

    # Calculate the rotation angle between the major axis and the target axis
    rotation_angle = (
        _signed_angle(
            major_axis_vector, target_axis_vector, plane_normal_vector
        )
        * 180
        / np.pi
    )

    # respects the right-hand rule wrt the corresponding normal vector
    rotation_plane_indices = {
        "XY": (-1, -2),
        "XZ": (-3, -1),
        "YZ": (-2, -3),
    }[rotation_plane]

    return rotation_angle, rotation_plane_indices



import glob
import json
import math
import os
from pathlib import Path
from xml.dom import minidom
import tifffile
import matplotlib.pyplot as plt
import napari
import numpy as np
import registrationtools
import io
from scipy.optimize import linear_sum_assignment
from scipy.spatial.transform import Rotation
from skimage.measure import regionprops
from skimage import io
import transforms3d._gohlketransforms as tg

def extract_positions(path_positions: str):
    """
    Extract the X and Y positions of the objects from the xml file (saved during the acquisition) and put them in a list
    Parameters
    ----------
    path_positions : str
        path to the xml file
    """
    print(path_positions)
    positions = minidom.parse(str(path_positions)) #need to convert the path to a string, otherwise minidom does not work for Windows paths
    pos_x = positions.getElementsByTagName("dXPosition")
    pos_y = positions.getElementsByTagName("dYPosition")

    xpos = np.zeros(pos_x.length)
    ypos = np.zeros(pos_y.length)

    for i in range(pos_y.length):
        xpos[i] = pos_x[i].attributes["value"].value
        ypos[i] = pos_y[i].attributes["value"].value

    return (xpos, ypos)

def plot_positions(path_ref_positions: str, path_float_positions: str):
    """
    Plot the positions of the objects from the xml file, to visualize and manually check what number have to be associated
    Parameters
    ----------
    path_ref_positions : str
        path to the xml file for the reference view
    path_float_positions : str
        path to the xml file for the floating view

    """
    (xpos_ref, ypos_ref) = extract_positions(path_ref_positions)
    (xpos_float, ypos_float) = extract_positions(path_float_positions)

    list_number_ref = [
        i + 1 for i in range(len(xpos_ref))
    ]  # the objects will have an index going from 1 to the total number, instead of having the id chosen during the aquisition. Problem ?
    fig, ax = plt.subplots()
    ax.scatter(xpos_ref, ypos_ref, label="ref")

    for i, txt in enumerate(list_number_ref):
        ax.annotate(txt, (xpos_ref[i], ypos_ref[i]))

    list_number_float = [i + 1 for i in range(len(xpos_float))]
    ax.scatter(xpos_float, ypos_float, label="float")

    for i, txt in enumerate(list_number_float):
        ax.annotate(txt, (xpos_float[i], ypos_float[i]))

    plt.legend()

def associate_positions(path_ref_positions: str, path_float_positions: str):
    """
    Associate the objects from the reference view with the objects from the floating view, by solving a linear sum assignement between the two distribution
    Parameters
    ----------
    path_ref_positions : str
        path to the xml file for the ref view
    path_float_positions : str
        path to the xml file for the float view

    Returns
    -------
    2 lists of indices, ref then float, sorted according to the asssgnement

    """
    
    (xpos_ref, ypos_ref) = extract_positions(path_ref_positions)
    (xpos_float, ypos_float) = extract_positions(path_float_positions)
    ypos_float=-ypos_float
    cost = np.zeros((len(xpos_ref), len(xpos_float)))
    for i in range(len(xpos_ref)):
        for j in range(len(xpos_float)):
            cost[i, j] = math.sqrt(
                math.pow(xpos_ref[i] - xpos_float[j], 2)
                + math.pow(ypos_ref[i] - ypos_float[j], 2)
            )
    row_ind, col_ind = linear_sum_assignment(cost)
    list_row = [int(i) for i in list(row_ind + 1)]
    list_col = [int(i) for i in list(col_ind + 1)]
    return (list_row,list_col)

def create_folders(
    folder_experiment: str,
    list_ref: list,
    list_float: list,
    channels: list,
    folder_output: str = "",
):
    """
    Creates the folders to save the registered images, and save the channels separately

    Parameters
    ----------
    folder_experiment : str
        path to the main folder
    list_ref : list
        list of the refrence images
    list_float : list
        list of the floating images corresponding to the reference images
    channels : list
        list of the names of the channels
    folder_output : str, optional
        path to the output folder, if None, will be the same as the experiment folder
    """

    if folder_output == "":
        folder_output = folder_experiment

    for ind_g in range(
        len(list_ref)
    ):  # for each sample, creates a dedicated folder and save the channels separately
        filename_ref = list_ref[ind_g]
        filename_float = list_float[ind_g]
        folder_sample = Path(folder_experiment) / filename_ref

        # creates paths for the output files
        os.mkdir(os.path.join(folder_experiment, filename_ref))
        os.mkdir(os.path.join(folder_sample, "trsf"))
        os.mkdir(os.path.join(folder_sample, "raw"))
        os.mkdir(os.path.join(folder_sample, "registered"))
        os.mkdir(os.path.join(folder_sample, "fused"))

        image_ref = io.imread(
            Path(folder_experiment) / f"{filename_ref}.tif"
        )
        image_float = io.imread(
            Path(folder_experiment) / f"{filename_float}.tif"
        )
        if len(channels)>1 :
            for ind_ch, ch in enumerate(channels):
                imref = (image_ref[:, :, :,ind_ch])
                imfloat = (image_float[:,:,:,ind_ch])
                io.imsave(
                    Path(folder_sample) / "raw" / f"{filename_ref}_{ch}.tif",
                    imref, ##CAREFUL needs to be float32 or uint16 orint16 otherwise the blockmatching does not compute/save the result
                )  # ,imagej=True, metadata={'axes': 'TZYX'})
                io.imsave(
                    Path(folder_sample) / "raw" / f"{filename_float}_{ch}.tif",
                    imfloat,
                )  # ,imagej=True, metadata={'axes': 'TZYX'})>
        else :
            io.imsave(
                Path(folder_sample) / "raw" / f"{filename_ref}_{channels[0]}.tif",
                image_ref, ##CAREFUL needs to be float32 or uint16 orint16 otherwise the blockmatching does not compute/save the result
            )
            io.imsave(
                Path(folder_sample) / "raw" / f"{filename_float}_{channels[0]}.tif",
                image_float,
            )  

def transformation_from_plugin(path_json: str,scale: tuple = (1, 1, 1)):
    """
    Extract the transformation from the json file saved by the plugin
    Parameters
    ----------
    path_json : str
        path to the json file
    scale : tuple, optional
        scale of the image, by default (1,1,1)
    """
    with open(path_json) as f:
        data = json.load(f)
    rot_z = data["rot_z"]
    rot_y = data["rot_y"]
    rot_x = data["rot_x"]
    trans_z = data["trans_z"]
    trans_y = data["trans_y"]
    trans_x = data["trans_x"]
    #in the napari plugin, the translation computed is the one after the rotation, so we need to set the translation trans1 before rotation to 0
    init_trsfs = list_init_trsf(trans1=[0, 0, 0], trans2=[trans_z, trans_y, trans_x], rot=[rot_x, rot_y, rot_z])
    return (init_trsfs)

def manual_registration_fct(
    reference_landmarks, floating_landmarks, scale: tuple = (1, 1, 1)
):
    # stolen from https://github.com/nghiaho12/rigid_transform_3D/blob/master/rigid_transform_3D.py
    """
    Finds the transformation between 2 sets of labels in 3D.
    If the automatic registration can't find an accurate transformation, 2 options :
    1) create landmarks layers using the napari plugin manual-registration, and save the values of rotation and transaltion into a json file that you will give as the argument "input_init_trsf_from_plugin" to the function register
    2) using another software to create 2 arays of the same size as your original images (without channels) and labels of the same values fot the 2 views. Save these arrays as tif files, and give them as arguments to this function.

    Parameters
    ----------
    reference_landmarks : np.array
        Landmarks for the reference image
    floating_landmarks : np.array
        Landmarks for the floating image

    Returns
    -------
    translation and rotation to apply to the floating image to register it on the reference image
    In the following order : rotation, translation1, translation2 with translation1 the translation to apply before rotation and translation2 the translation to apply after rotation
    """

    rg_ref = regionprops(reference_landmarks)
    centroids_ref = np.array([prop.centroid for prop in rg_ref])
    centroids_ref = (centroids_ref * scale).T

    rg_float = regionprops(floating_landmarks)
    centroids_float = np.array([prop.centroid for prop in rg_float])
    centroids_float = (centroids_float * scale).T

    assert centroids_ref.shape == centroids_float.shape

    num_rows, num_cols = centroids_ref.shape
    if num_rows != 3:
        raise Exception(
            f"matrix centroids_ref is not 3xN, it is {num_rows}x{num_cols}"
        )

    num_rows, num_cols = centroids_float.shape
    if num_rows != 3:
        raise Exception(
            f"matrix centroids_float is not 3xN, it is {num_rows}x{num_cols}"
        )

    # find mean column wise
    centermass_ref = np.mean(centroids_ref, axis=1)
    centermass_float = np.mean(centroids_float, axis=1)

    # subtract mean
    centroids_ref_centered = centroids_ref - centermass_ref.reshape(3, 1)
    centroids_float_centered = centroids_float - centermass_float.reshape(3, 1)

    H = centroids_ref_centered @ np.transpose(centroids_float_centered)

    # find rotation
    U, S, Vt = np.linalg.svd(H)
    R = Vt.T @ U.T

    # special reflection case
    if np.linalg.det(R) < 0:
        print("det(R) < R, reflection detected!, correcting for it ...")
        Vt[2, :] *= -1
        R = Vt.T @ U.T
    t = -R @ centermass_ref + centermass_float

    trsf = np.identity(4)
    trsf[0:3, 0:3] = R
    trsf[0:3, 3] = t
    trsf = np.linalg.lstsq(trsf, np.identity(4))[0]
    np.savetxt("trsf.txt", trsf)

    rotation = Rotation.from_matrix(R)
    rotation_angles = rotation.as_euler("xyz", degrees=True)

    # translation1 is the translation that will center the floating landmarks on the center of the image.
    # Then, we will apply the rotation with respect to the center (the center of mass stays at the center))
    # Afterwards, we apply translation2 to center the floating landmarks that was located at the center, on the reference position.

    center_image = (np.array(reference_landmarks.shape) / 2) * scale
    translation1 = center_image - centermass_float
    translation2 = centermass_ref - center_image
    return (list(rotation_angles), list(translation1), list(translation2)) #into list to be able to copy paste directly into the registration function

def list_init_trsf(trans1,trans2,rot):
    """
    Transforms the individual values of rotation and translation into a list of transformations to apply in the right order. This list can be given as the parameter other_trsf to the function register
    Parameters
    ----------
    trans1 : list
        list of the translations to apply to the floating image BEFORE ROTATION
    trans2 : list   
        list of the translations to apply to the floating image AFTER ROTATION
    rot : list  
        list of the rotations to apply to the floating image
    """

    trans2_in_um=trans2
    trans1_in_um=trans1
    init_trsf = [
            [
                "trans",
                "X",
                -trans1_in_um[2],
                "trans",
                "Y",
                -trans1_in_um[1],
                "trans",
                "Z",
                -trans1_in_um[0],
                "rot",
                "X",
                -rot[0],
                "rot",
                "Y",
                -rot[1],
                "rot",
                "Z",
                -rot[2],
                "trans",
                "X",
                -trans2_in_um[2],
                "trans",
                "Y",
                -trans2_in_um[1],
                "trans",
                "Z",
                -trans2_in_um[0],
            ]
        ]
    return(init_trsf)

def compute_transformation_from_trsf_files(path_trsf:str):
        
    f = open(Path(path_trsf) / "A1-rigid.trsf", 'r')

    # Read and ignore header lines (NECESSARY)
    header1 = f.readline()
    header2 = f.readline()

    matrix=np.zeros((4,4))

    index=0;
    for line in f:
        line = line.strip()
        columns = line.split()
        if index<4:
            matrix[index,:] = np.array(columns, dtype=float)
        index=index+1
        
    scale, shear, angles, trans, persp = tg.decompose_matrix(matrix)
    angles_deg=[math.degrees(angles[0]), math.degrees(angles[1]), math.degrees(angles[2])]
    print("rotation angles (X,Y,Z) in deg:",angles_deg)
    print("translation vector:",trans)
def register(
    path_data: str,
    path_transformation: str,
    path_registered_data: str,
    reference_image: str,
    floating_image: str,
    input_voxel: list = [1, 1, 1],
    output_voxel: list = [1, 1, 1],
    compute_trsf: int = 1,
    trans1: list = None,
    rot: list = None,
    trans2: list = None,
    other_trsf: list = None,
    input_init_trsf_from_plugin: str = "",
    test_init: int = 0,
    trsf_type: str = "rigid",
    depth: int = 3,
    bbox: int = 1,
    image_interpolation: str = "linear",
    padding: int = 0,
    save_json: str = "",
    ordered_init_trsfs: bool = True,
):

    # Register the two sides of the sample, using the previously computed transformation (if any) or computing a new one
    # """
    # Parameters
    # ----------
    # path_data : str
    #     path to the raw images. In the folder structure : folder_experiment/sample_id/raw
    # path_transformation : str
    #     path to the folder where the transformations files are saved. In the folder structure : folder_experiment/sample_id/trsf
    # path_registered_data : str
    #     path where the registered images will be saved. In the folder structure : folder_experiment/sample_id/registered
    # reference_image : str
    #     name of the reference image, the 'fixed' one
    # floating_image : str
    #     name of the floating image, the one that will be registered onto the reference image
    # input_voxel : tuple, optional
    #     voxel size of the input image, by default [1,1,1]
    # output_voxel : tuple, optional
    #     voxel size of the output image, by default [1,1,1]. Can be different from the input voxel size.
    # compute_trsf : int, optional
    #     1 if the transformation has to be computed, 0 if it already exists. If you have multiple channels of the same image, it is recommended to pick one expressed homogeneously as the reference, register this channel using compute_trsf=1.
    #     Then you can use compute_trsf=0, the algo will find the pre-existing transformation to register the other channels onto the reference channel.
    # rot : list, optional
    #     list of the rotations to apply to the floating image, by default [0,0,0]
    # trans1 : list, optional
    #     list of the translations to apply to the floating image BEFORE ROTATION, by default [0,0,0]
    # trans2 : list, optional
    #     list of the translations to apply to the floating image AFTER ROTATION, by default [0,0,0]
    # other_trsf : list, optional.
    #     list of transformations to apply to the floating image, by default [], if you want give direclty your transformations.
    #     If this argument is not None, the value of the parameters rot, trans1, trans2 will be ignored.
    #     You can use flipping (flip), rotations (rot), translation (trans). Specify the axis X,Y or Z after the sample_id of the transformation. For rotations and translation, precise the value (angle or distance) after the axis.
    #     IMPORTANT : Needs 2 sets of brackets.
    #     If order_init_trsfs is True, the transformations will be applied in the order they are given in the list init_trsfs.
    #     # Example : [['flip', 'Z', 'trans', 'Z', -10,'trans','Y',100,'rot','X',-29,'rot','Y',41,'rot','Z',-2]]
    # input_init_trsf_from_plugin : str, optional
    #     path to the json file saved by the plugin, containing the transformation. If this parameter is not None, the transformation will be extracted from the json file no matter the parameters rot, trans1, trans2 or other_trsf
    # test_init : int, optional
    #     1 if we apply only the initial transformation. 0 if we apply the initial trsf + blockmatching algo. By default 0. 
    # trsf_type : str, optional
    #     type of transformation to compute : rigid, affine. By default rigid.
    # depth : int, optional
    #     depth of the registration, by default 3
    # bbox : int, optional
    #     1 if the bounding box of the original image can extend, 0 if not.
    # image_interpolation : str, optional
    #     type of interpolation to apply to the image, by default 'linear'
    # padding : int, optional
    #     padding to apply to the image, by default 0
    # save_json : str
    #     if not None, save the parameters of that function in a json file at the path given.
    #     IMPORTANT : saves the parameters of the function, not the parameters of the registration, in particular the initial transformations only. To save the actual transformations, use the function "compute_transformation_from_trsf_files"
    # ordered_init_trsfs : bool, optional
    #     if True, the transformations will be applied in the order they are given in the list init_trsfs.

    # """
    if rot is None:
        rot = [0, 0, 0]
    if trans1 is None:
        trans1 = [0, 0, 0]
    if trans2 is None:
        trans2 = [0, 0, 0]

    if input_init_trsf_from_plugin != "":
    #if the user input the json file from the plugin, the json file is used as init trsf no matter the parameters rot, trans1, trans 2 or other_trsf
        init_trsfs=transformation_from_plugin(input_init_trsf_from_plugin,scale=input_voxel)
    elif input_init_trsf_from_plugin == "" and other_trsf is not None:
        print('other_trsf',other_trsf)
    # if the user defined the initial transformation with the parameter 'other_trsf' (list of transformations he wants), then we use this transformation as init_trsfs
        init_trsfs = other_trsf
    else:  # is the user did not precise other_trsf or any json file, then we use the parameters rot, trans1 and trans2 to "build" init_trsfs.
        init_trsfs = list_init_trsf(trans1=trans1,trans2=trans2,rot=rot)
    data = {
        "path_to_data": str(path_data),
        "ref_im": reference_image,
        "flo_ims": [floating_image],
        "compute_trsf": compute_trsf,
        "init_trsfs": init_trsfs,
        "trsf_paths": [str(path_transformation)],
        "trsf_types": [trsf_type],
        "ref_voxel": input_voxel,
        "flo_voxels": [input_voxel],
        "out_voxel": output_voxel,
        "test_init": test_init,
        "apply_trsf": 1,
        "out_pattern": str(path_registered_data),
        "begin": 1,
        "end": 1,
        "bbox_out": bbox,
        "image_interpolation": image_interpolation,
        "padding": padding,
        "registration_depth": depth,
        "ordered_init_trsfs": ordered_init_trsfs,
    }
    print(data)

    if save_json != "":
        json_string = json.dumps(data, indent=4)
        with open(Path(save_json) / "parameters.json", "w") as outfile:
            outfile.write(json_string)
    tr = registrationtools.SpatialRegistration(data)
    tr.run_trsf()

def check_napari(
    folder: str,
    reference_image,
    floating_image,
    additional_images: list = None,
    names_additional_images: list = None,
    scale: tuple = (1, 1, 1),
    labels: bool = False,
):
    """
    Opens the registered images in napari , to check how they overlap

    Parameters
    ----------
    folder : str
        path to the main folder. The floating image is always saved in the folder 'registered' but not the reference image.
    reference_image : str or ndarray (can be the name of the array in path_data or the array itself)
        name of the reference image, the 'fixed' one
    floating_image : str or ndarray (can be the name of the array in path_data or the array itself)
        name of the floating image, the one that will be registered onto the reference image
    additional_images : list of ndarray, optional
        list of additional images to add to the viewer, by default []
    scale : tuple, optional
        scale of the image (should be the same ), by default (1,1,1)
    labels : bool, optional
        if True, the images are considered as labels, by default False

    """

    viewer = napari.Viewer()
    if isinstance(reference_image, (str | Path)):
        if os.path.exists(
            Path(folder) / "registered" / reference_image
        ):  # if the bounding box or voxel size has changed, then the reference image has been saved as an output.
            ref_im = io.imread(
                Path(folder) / "registered" / reference_image
            )
        else:  # if the bbox did not change, the reference image has not changed and is found in the raw folder only.
            ref_im = io.imread(Path(folder) / "raw" / reference_image)
    else:
        ref_im = reference_image

    if isinstance(floating_image, (str | Path)):
        float_im = io.imread(
            Path(folder) / "registered" / floating_image
        )
    else:
        float_im = floating_image

    if not labels:
        viewer.add_image(
            ref_im, colormap="cyan", name="reference_image", scale=scale
        )
        viewer.add_image(
            float_im,
            colormap="red",
            blending="additive",
            name="floating_image",
            scale=scale,
        )
        if additional_images is not None:
            for num_im, im in enumerate(additional_images):
                viewer.add_image(
                    im,
                    colormap="green",
                    blending="additive",
                    name=names_additional_images[num_im],
                    scale=scale,
                )
    if labels:
        viewer.add_labels(ref_im, name="reference_image", scale=scale)
        viewer.add_labels(float_im, name="floating_image", scale=scale)
        if additional_images is not None:
            for num_im, im in enumerate(additional_images):
                name = names_additional_images[num_im]
                viewer.add_labels(im, name=str(name), scale=scale)

    napari.run()

def sigmoid(z, z0, p):
    """
    Sigmoid function, to weight the images by the distance to the edges
    z0 gives the middle of the sigmoid, at which weight=0.5
    p gives the slope. 5 corresponds to a low slope, wide fusion width and 25 to a strong slope, very thin fusion width.
    """
    return 1 / (1 + np.exp(-p * (z - z0)))

def fuse_sides(
    path_registered_data: str,
    reference_image_reg: str,
    floating_image_reg: str,
    folder_output: str = "",
    name_output: str = "fusion",
    slope_coeff: int = 20,
    axis: int = 0,
    return_image=False
):
    """
    Fuse the two sides of the sample, using the previously registered images

    Parameters
    ----------
    path_registered_data : str
        path to the registered images
    reference_image_reg : str
        name of the reference image, the 'fixed' one. Can be its name in path_registered_data or the array itself.
    floating_image_reg : str
        name of the floating image, the one that will be registered onto the reference image. Can be its name in path_registered_data or the array itself
    folder_output : str
        path to the folder where the fused image will be saved
    name_output : str, optional
    slope_coeff : int, optional
        coefficient to apply to the sigmoid function, by default 20.
        name of the output which is the fused image , by default 'fusion'
    axis : int, optional
        axis along which the fusion is done, by default 0 (z axis)
    return_image : bool, optional
        if True, returns the fused image, by default False
    """

    ###Takes the previously saved images (two registered sides)
    if isinstance(reference_image_reg, (str | Path)): #if the reference image is a string, then we load the image from the path
        ref_image = io.imread(
            Path(path_registered_data) / reference_image_reg
        )
    #else its the image itself
    if isinstance(floating_image_reg, (str | Path)):
        float_image = io.imread(
            Path(path_registered_data) / floating_image_reg
        )

    dtype_input=float_image.dtype
    mask_r = ref_image > 0
    mask_f = float_image > 0
    mask_fused = mask_r & mask_f
    cumsum = np.cumsum(mask_fused, axis=axis)
    # we take the cumulative sum  along the fusion axis, this will give me a linear weight.
    cumsum_normalized = cumsum / np.max(cumsum)

    # apply a sigmoid function to the linear weights, to get a smooth transition between the two sides
    w2 = sigmoid(cumsum_normalized, z0=0.5, p=slope_coeff)
    w1 = 1 - w2

    fusion = (ref_image * w1 + float_image * w2)

    if return_image:
        return fusion
    io.imsave(Path(folder_output) / name_output, fusion.astype(dtype_input))

def write_hyperstacks(
    path: str, sample_id: str, channels: list, return_image=False,dtype=np.float32):
    """
    Writes the hyperstacks, by stacking the channels of the registered images

    Parameters
    ----------
    path : str
        path to the registered images
    sample_id : str
        name of the sample
    channels : list
        list of the names of the channels
    return_image : bool, optional
        if True, returns the hyperstack, by default False
    """

    image = io.imread(
        Path(path) / f"{sample_id}_{channels[0]}.tif")  # reading one image just to extract the shape of the image and initialize the hyperstack
    (z, y, x) = image.shape
    new_image = np.zeros((z ,len(channels),y, x))
    for ch in range(len(channels)):
        one_channel = io.imread(
            Path(path) / f"{sample_id}_{channels[ch]}.tif")
        print(one_channel.shape)
        new_image[:, ch,:, :] = one_channel

    if return_image:
        return new_image

    tifffile.imwrite(Path(path) / f"{sample_id}_registered.tif", new_image.astype(dtype), imagej=True, compression=('zlib', 1))#float16 not compatible with Fiji


def add_centermass(landmarks, radius: int = 10, centermass_label: int = 10):
    """
    For debug purposes : if the registration does not work, you can use this function on the landmarks image to check if the center of mass is correcly aligned after the rotation.

    Parameters
    ----------
    landmarks : np.array
        array containing the landmarks (can be float or reference)
    radius : int, optional
        radius of the sphere to add around the center of mass, by default 10
    centermass_label : int, optional
        value of the label of the center of mass, by default 10 (arbitrary)
    """

    rg = regionprops(landmarks)
    centroids = np.array([prop.centroid for prop in rg]).T
    z, y, x = np.mean(centroids, axis=1)
    landmarks[
        int(z) - radius : int(z) + radius,
        int(y) - radius : int(y) + radius,
        int(x) - radius : int(x) + radius,
    ] = centermass_label
    return landmarks

def remove_previous_files(path):
    filelist = glob.glob(os.path.join(path, "*.tif"))
    for f in filelist:
        os.remove(f)



import numpy as np



def tensors_to_napari_vectors(
        sparse_tensors: np.ndarray, 
        is_inertia_tensor: bool, 
        volumes: np.ndarray = None, 
        return_angles: bool = False
    ):
    """
    Extracts the largest eigenvalue and eigenvector from a set of sparse tensors.
    Returns the product of the eigenvalues and eigenvectors as a set of vector
    array ready to be plotted in napari.

    Parameters
    ----------
    sparse_tensors : np.ndarray
        A 2D array of shape (n_tensors, dim_space + dim_space**2) where the first
        dim_space columns are the positions at which the tensors are defined and
        the next dim_space**2 columns are the tensors values.
    is_inertia_tensor : bool
        If True, the function will process the eigenvalues of the tensors in a
        special manner to extract the principal lengths. Otherwise, the tensors
        are simply diagonalized.
    volumes : np.ndarray, optional
        The volumes of the regions associated with the tensors. It is only useful
        (and necessary) to provide it wien is_inertia_tensor is True.
    return_angles : bool, optional
        If True, the function will return the angles of the principal vectors with
        respect to the X-axis.

    Returns
    -------
    napari_vectors : np.ndarray
        A 3D array V of shape (n_tensors, 2, dim_space) where V[i,0] is the
        position at which the vector is defined and V[i,1] is the vector itself.
    angles : np.ndarray
        The angles of the principal vectors with respect to the X-axis. Only
        returned if return_angles is True.
    """

    positions = sparse_tensors[:, :3]
    tensors = sparse_tensors[:, 3:]
    tensors = tensors.reshape(-1, 3, 3)

    # diagonalize the tensors
    eigen_values, principal_vectors = np.linalg.eigh(tensors)

    if is_inertia_tensor:
        # the principal lengths are a mix of the eigenvalues and the volumes
        axis_decoupling_matrix = np.ones((3,3))/2 - np.eye(3)
        eigen_values = np.sqrt(
            np.einsum('ij,lj->li', axis_decoupling_matrix, eigen_values)
        )
        eigen_values = eigen_values * np.sqrt(5 / volumes.reshape(-1, 1, 1))
    
    # WARNING: the scipy documentation is wrong, the eigenvectors are in the columns
    # not the lines, i.e principal_vectors[:,:,0] is the 0th principal vector
    principal_vectors = principal_vectors.transpose(0,2,1)

    napari_vectors = np.zeros((len(positions), 2, 3))
    indices_maxs = np.nanargmax(eigen_values, axis=1)
    principal_vectors = principal_vectors[np.arange(len(eigen_values)), indices_maxs]
    eigen_values = eigen_values[np.arange(len(eigen_values)), indices_maxs]
    napari_vectors[:,0] = positions
    napari_vectors[:,1] = principal_vectors * eigen_values.reshape(-1,1)

    if return_angles:
        angles = np.arctan2(*(napari_vectors[:,1, -2:].reshape(-1, 2).T))
        angles = np.arctan2(np.sin(angles), np.cos(angles))

        return napari_vectors, angles


    return napari_vectors


import matplotlib.pyplot as plt
import matplotlib
import numpy as np
from matplotlib.colors import LinearSegmentedColormap
from skimage.measure import regionprops
from skimage.filters import threshold_otsu

import tapenade.utils as utils


class SpatialCorrelationPlotter:
    def __init__(
        self,
        quantity_X: np.ndarray,
        quantity_Y: np.ndarray,
        mask: np.ndarray = None,
        labels: np.ndarray = None,
    ):
        """
        Initialize computation on two quantities X and Y before plotting
        a heatmap of their spatial correlation. If mask is provided, the
        computation will be restricted to the masked region. If labels are
        provided, the computation will be restricted to the labeled regions,
        regardless of the mask.

        Parameters:
        - quantity_X: numpy array representing the first quantity
        - quantity_Y: numpy array representing the second quantity
        - mask: optional numpy array, binary mask that will be applied to restrict
            the values of quantity_X and quantity_Y
        - labels: optional numpy array, labels of object instances (e.g nuclei) in
            which the the values of quantity_X and quantity_Y will be averaged.
            Each object instance will create a single point in the heatmap.
        """

        if labels is not None:
            if mask is None:
                props = regionprops(labels, intensity_image=quantity_X)
                quantity_X = np.array([prop.mean_intensity for prop in props])
                props = regionprops(labels, intensity_image=quantity_Y)
                quantity_Y = np.array([prop.mean_intensity for prop in props])
            else:
                props = regionprops(labels, intensity_image=quantity_X)
                quantity_X = np.array(
                    [
                        prop.mean_intensity
                        for prop in props
                        if np.any(mask[prop.slice][prop.image])
                    ]
                )
                props = regionprops(labels, intensity_image=quantity_Y)
                quantity_Y = np.array(
                    [
                        prop.mean_intensity
                        for prop in props
                        if np.any(mask[prop.slice][prop.image])
                    ]
                )
        elif mask is not None:
            quantity_X = quantity_X[mask]
            quantity_Y = quantity_Y[mask]
        else:
            quantity_X = quantity_X.ravel()
            quantity_Y = quantity_Y.ravel()

        self.quantity_X = quantity_X
        self.quantity_Y = quantity_Y
        self.mask = mask
        self.labels = labels

        self.xys = np.array([self.quantity_X, self.quantity_Y]).T

    def _normalize_quantity(self, quantity: np.ndarray):
        return (quantity - np.median(quantity)) / np.std(quantity)

    def get_heatmap_figure(
        self,
        bins: tuple = (40, 40),
        show_individual_cells: bool = False,
        show_linear_fit: bool = True,
        normalize_quantities: bool = False,
        extent_X: tuple = None,
        extent_Y: tuple = None,
        percentiles_X: tuple = (0, 100),
        percentiles_Y: tuple = (0, 100),
        log_scale_X: bool = False,
        log_scale_Y: bool = False,
        figsize: tuple = (7, 4),
        label_X: str = "X",
        label_Y: str = "Y",
        colormap: str = "plasma",
        sample_fraction: float = 0.005,
        display_quadrants: bool = False,
        fig_ax_tuple: tuple = None,
    ):
        """
        Create a heatmap of the spatial correlation between two quantities X and Y.

        Parameters:
        - bins: list of two integers, number of bins in each dimension
        - show_individual_cells: bool, if True and if labels was specified in __init__,
            individual cells will be displayed as a scatter plot on top of the heatmap.
        - show_linear_fit: bool, if True, a linear fit will be displayed on the heatmap.
        - normalize_quantities: bool, if True, the quantities will be normalized to have
            zero mean and unit standard deviation.
        - extent_X: tuple of two floats, the extent of the X axis
        - extent_Y: tuple of two floats, the extent of the Y axis
        - percentiles_X: tuple of two floats, only the values of X between these percentiles
            will be considered
        - percentiles_Y: tuple of two floats, the percentiles of Y between thse percentiles
            will be considered
        - log_scale_X: bool, if True, the X axis will be displayed in log scale
        - log_scale_Y: bool, if True, the Y axis will be displayed in log scale
        - figsize: tuple of two floats, the size of the figure
        - label_X: str, the label of the X axis
        - label_Y: str, the label of the Y axis
        - colormap: str, the name of the colormap
        - sample_fraction: float, if labels has not been specified in __init__, sample_fraction
            represents the fraction of the total number of values of X and Y that will be
            randomly selected to be used for the computation of the statistics.
        - display_quadrants: bool, if True, quadrants will be displayed as vertical and horizontal
            dashed lines at the Otsu thresholds of X and Y.
        - fig_ax_tuple: tuple of matplotlib figure and axis, if not None, the heatmap will be
            displayed on the provided figure and axis.

        Returns:
        - fig: matplotlib figure
        - ax: matplotlib axis

        """

        quantity_X = self.quantity_X.copy()
        quantity_Y = self.quantity_Y.copy()

        if normalize_quantities:
            quantity_Y = self._normalize_quantity(quantity_Y)
            quantity_X = self._normalize_quantity(quantity_X)
            log_scale_X = False
            log_scale_Y = False

        quantity_X, quantity_Y = utils.filter_percentiles(
            X=quantity_X,
            percentilesX=percentiles_X,
            Y=quantity_Y,
            percentilesY=percentiles_Y,
        )

        if extent_X is None:
            extent_X = (np.min(quantity_X), np.max(quantity_X))
        else:
            bins[0] = np.linspace(extent_X[0], extent_X[1], bins[0] + 1)
        if extent_Y is None:
            extent_Y = (np.min(quantity_Y), np.max(quantity_Y))
        else:
            bins[1] = np.linspace(extent_Y[0], extent_Y[1], bins[1] + 1)

        bins = list(bins)

        if fig_ax_tuple is None:
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig, ax = fig_ax_tuple

        if log_scale_X:
            bins[0] = np.logspace(
                np.log10(extent_X[0]), np.log10(extent_X[1]), bins[0]
            )
        if log_scale_Y:
            bins[1] = np.logspace(
                np.log10(extent_Y[0]), np.log10(extent_Y[1]), bins[1]
            )

        heatmap, xedges, yedges = np.histogram2d(
            quantity_X.ravel(),
            quantity_Y.ravel(),
            bins=tuple(bins),
        )

        extent = extent_X + extent_Y

        custom_cmap = self._get_custom_cmap(colormap)

        # Create heatmap
        ims = ax.imshow(
            heatmap.T,
            origin="lower",
            cmap=custom_cmap,
            extent=extent,
            aspect="auto",
            interpolation="none",
        )
        cbar = fig.colorbar(ims, ax=ax)
        cbar.set_label("Number of cells", fontsize=10)

        if self.labels is not None:
            sample_indices = None
            self._add_individual_scatter(
                ax,
                normalize_quantities,
                show_individual_cells,
                ims,
                heatmap,
                xedges,
                yedges,
                custom_cmap,
            )
        else:
            n = int(sample_fraction * len(self.quantity_X))
            sample_indices = np.random.choice(
                len(self.quantity_X), n, replace=False
            )

            self._add_individual_scatter(
                ax,
                normalize_quantities,
                show_individual_cells,
                ims,
                heatmap,
                xedges,
                yedges,
                custom_cmap,
                sample_indices=sample_indices,
            )

        if display_quadrants:
            self._display_quadrants(ax, quantity_X, quantity_Y)

        if show_linear_fit:
            self._add_linear_fit(ax, quantity_X, quantity_Y, xedges)

        if not log_scale_X and not log_scale_Y:
            ax.ticklabel_format(style="sci", axis="both", scilimits=(0, 0))

        self._format_figure(
            fig,
            ax,
            log_scale_X,
            log_scale_Y,
            label_X,
            label_Y,
            extent_X,
            extent_Y,
        )

        return fig, ax

    def _get_custom_cmap(self, colormap: str):
        # changing colormap to get white for 0 values
        inferno_cmap = plt.cm.get_cmap(colormap)
        inferno_colors = inferno_cmap(np.linspace(0, 1, 256))
        inferno_colors[0] = (1, 1, 1, 1)  # (R, G, B, Alpha)
        custom_cmap = LinearSegmentedColormap.from_list(
            "custom_inferno", inferno_colors
        )

        return custom_cmap

    def _add_individual_scatter(
        self,
        ax: plt.Axes,
        normalize_quantities: bool,
        show_individual_cells: bool,
        ims: plt.imshow,
        heatmap: np.ndarray,
        xedges: np.ndarray,
        yedges: np.ndarray,
        colormap: str,
        sample_indices: list = None,
    ):

        if sample_indices is None:
            quantity_X = self.quantity_X
            quantity_Y = self.quantity_Y
        else:
            quantity_X = self.quantity_X[sample_indices]
            quantity_Y = self.quantity_Y[sample_indices]

        if normalize_quantities:
            quantity_X = self._normalize_quantity(quantity_X)
            quantity_Y = self._normalize_quantity(quantity_Y)

        indX = np.clip(
            np.digitize(quantity_X, xedges) - 1, 0, heatmap.shape[0] - 1
        )
        indY = np.clip(
            np.digitize(quantity_Y, yedges) - 1, 0, heatmap.shape[1] - 1
        )

        vals = heatmap[indX, indY]

        scatter = ax.scatter(
            quantity_X,
            quantity_Y,
            c=vals,
            cmap=colormap,
            s=3,
            vmin=heatmap.min(),
            vmax=heatmap.max(),
        )

        if show_individual_cells:
            # change alpha of imshow
            ims.set_alpha(0.25)
        else:
            # change alpha of scatter
            scatter.set_alpha(0)

    def _add_linear_fit(
        self,
        ax: plt.Axes,
        quantity_X: np.ndarray,
        quantity_Y: np.ndarray,
        xedges: np.ndarray,
    ):
        intercept, slope, r2 = utils.linear_fit(
            x=quantity_X, y=quantity_Y, return_r2=True
        )

        x = np.array([xedges[0], xedges[-1]])
        lin_fit = intercept + slope * x
        ax.plot(
            x,
            lin_fit,
            c="red",
            lw=3,
            label=f"Y = {slope:.1E} X + {intercept:.1E},\nR²={r2:.2f}",
        )
        ax.legend(fontsize=10)

    def _display_quadrants(
        self, ax: plt.Axes, quantity_X: np.ndarray, quantity_Y: np.ndarray
    ):

        threshold_X = threshold_otsu(quantity_X)
        threshold_Y = threshold_otsu(quantity_Y)

        ax.axvline(threshold_X, color="black", linestyle="--")
        ax.axhline(threshold_Y, color="black", linestyle="--")

    def _format_figure(
        self,
        fig: plt.Figure,
        ax: plt.Axes,
        log_scale_X: bool,
        log_scale_Y: bool,
        label_X: str,
        label_Y: str,
        extent_X: tuple,
        extent_Y: tuple,
    ):

        if log_scale_X:
            ax.set_xscale("log")
        if log_scale_Y:
            ax.set_yscale("log")

        ax.set_xlim(extent_X)
        ax.set_ylim(extent_Y)

        xlim = list(ax.get_xlim())
        ylim = list(ax.get_ylim())

        x_diff = xlim[1] - xlim[0]
        xlim[0] = xlim[0] - 0.025 * x_diff
        xlim[1] = xlim[1] + 0.025 * x_diff
        ax.set_xlim(xlim)

        y_diff = ylim[1] - ylim[0]
        ylim[0] = ylim[0] - 0.025 * y_diff
        ylim[1] = ylim[1] + 0.025 * y_diff
        ax.set_ylim(ylim)

        if not log_scale_X and not log_scale_Y:
            ax.ticklabel_format(style="sci", axis="both", scilimits=(0, 0))

        matplotlib.rc("ytick", labelsize=10)
        matplotlib.rc("xtick", labelsize=10)

        ax.set_xlabel(label_X, fontsize=12)
        ax.set_ylabel(label_Y, fontsize=12)

        fig.tight_layout()

    def plot_heatmap(
        self,
        bins: tuple = (20, 20),
        show_individual_cells: bool = False,
        show_linear_fit: bool = True,
        extent_X: tuple = None,
        extent_Y: tuple = None,
        extent_percentiles: tuple = None,
        log_scale_X: bool = False,
        log_scale_Y: bool = False,
        display_quadrants: bool = False,
    ):
        """
        Create and plot a heatmap of the spatial correlation between two
        quantities X and Y.

        See get_heatmap_figure for the description of the parameters.
        """

        fig, ax = self.get_heatmap_figure(
            bins=bins,
            show_individual_cells=show_individual_cells,
            show_linear_fit=show_linear_fit,
            extent_X=extent_X,
            extent_Y=extent_Y,
            extent_percentiles=extent_percentiles,
            log_scale_X=log_scale_X,
            log_scale_Y=log_scale_Y,
            display_quadrants=display_quadrants,
        )

        fig.show()



import numpy as np
from scipy.ndimage import zoom


def _change_array_pixelsize(
    array: np.ndarray,
    input_pixelsize: tuple[float, float, float] = (1, 1, 1),
    output_pixelsize: tuple[float, float, float] = (1, 1, 1),
    order: int = 1,
):
    """
    Make the input array isotropic by resampling it to the target spacing.

    Parameters:
    - mask (ndarray): The mask indicating the valid region.
    - image (ndarray): The image to be resampled.
    - labels (ndarray): The segmentation labels.
    - input_pixelsize (tuple): The input pixel size (e.g in microns).
    - output_pixelsize (tuple): The output pixel size (e.g in microns).
    - order (int): The order of the spline interpolation.

    Returns:
        ndarray: The resampled mask, image, or labels.
    """

    reshape_factors = np.array(input_pixelsize) / np.array(output_pixelsize)

    array_isotropic = zoom(array, reshape_factors, order=order)

    return array_isotropic



import numpy as np
from scipy.optimize import least_squares
from scipy.stats import linregress


def filter_percentiles(
    X, percentilesX: tuple = (1, 99), Y=None, percentilesY: tuple = None
):

    if Y is None:

        down, up = percentilesX

        percentile_down = np.percentile(X, down)
        percentile_up = np.percentile(X, up)

        mask = np.logical_and(percentile_down < X, percentile_up > X)

        return X[mask]

    else:

        downX, upX = percentilesX

        if percentilesY is None:
            downY, upY = percentilesX
        else:
            downY, upY = percentilesY

        percentile_downX = np.percentile(X, downX)
        percentile_downY = np.percentile(Y, downY)

        percentile_upX = np.percentile(X, upX)
        percentile_upY = np.percentile(Y, upY)

        maskX = np.logical_and(percentile_downX <= X, percentile_upX >= X)
        maskY = np.logical_and(percentile_downY <= Y, percentile_upY >= Y)

        mask = np.logical_and(maskX, maskY)

        return X[mask], Y[mask]


def linear_fit(
    x,
    y,
    robust: bool = False,
    return_r2: bool = False,
    robust_params_init: tuple = None,
    robust_f_scale: float = None,
):

    if not robust:
        res = linregress(x, y)

        if return_r2:
            return res.intercept, res.slope, res.rvalue**2
        else:
            return res.intercept, res.slope

    else:

        def f(params, x, y):
            return params[0] + params[1] * x - y

        if robust_params_init is None:
            robust_params_init = np.ones(2)

        res_robust = least_squares(
            f,
            robust_params_init,
            args=(x, y),
            loss="soft_l1",
            f_scale=robust_f_scale,
        )

        if return_r2:
            raise NotImplementedError
        else:
            return res_robust.x[0], res_robust.x[1]



from typing import Union

import numba
import numpy as np
from scipy.ndimage import gaussian_filter as scipy_gaussian
from scipy.spatial import KDTree


def _masked_smooth_gaussian(
    array: np.ndarray,
    sigmas: Union[float, tuple],
    mask: np.ndarray = None,
    mask_for_volume: np.ndarray = None,
):
    """
    If 'mask' is specified, the convolution will not take the
    masked value into account.
    """

    sigmas = sigmas if isinstance(sigmas, int | float) else np.array(sigmas)

    if mask is None:
        return scipy_gaussian(array, sigmas, mode="nearest", cval=0.0)
    else:

        mask = mask.astype(bool)

        if mask_for_volume is None:

            smooth_array = scipy_gaussian(
                np.where(mask, array.astype(np.float32), 0.0),
                sigmas,
                mode="constant",
                cval=0.0,
                truncate=3.0,
            )

            # calculate renormalization factor for masked gaussian (the 'effective'
            # volume of the gaussian kernel taking the mask into account)
            effective_volume = scipy_gaussian(
                mask.astype(np.float32),
                sigmas,
                mode="constant",
                cval=0.0,
                truncate=3.0,
            )

        else:
            mask_for_volume = mask_for_volume.astype(bool)

            smooth_array = scipy_gaussian(
                np.where(mask_for_volume, array.astype(np.float32), 0.0),
                sigmas,
                mode="constant",
                cval=0.0,
                truncate=3.0,
            )

            # calculate renormalization factor for masked gaussian (the 'effective'
            # volume of the gaussian kernel taking the mask into account)
            effective_volume = scipy_gaussian(
                mask_for_volume.astype(np.float32),
                sigmas,
                mode="constant",
                cval=0.0,
                truncate=3.0,
            )

        smooth_array = np.where(
            mask,
            np.divide(smooth_array, effective_volume, where=mask),
            0.0,
        )

        return smooth_array


@numba.jit(nopython=True, fastmath=True)
def _loop_numba(cols, values, dists, n_points, dim_points):
    """
    Computes the actual smoothing kernel operation from the old to the new positions.
    NEVER SET PARALLEL=True IN NUMBA DECORATOR, IT WILL BREAK THE CODE.
    """

    smoothed_points = np.zeros((n_points, dim_points))
    normalization_values = np.zeros((n_points, 1))
    # normalization_values = np.ones((n_points,1))  # * 0.1

    for i in range(len(values)):
        exp_dist = np.exp(-dists[i] ** 2 / 2)
        smoothed_points[cols[i]] += values[i] * exp_dist
        normalization_values[cols[i]] += exp_dist

    for j in range(n_points):
        if normalization_values[j] > 0:
            smoothed_points[j] /= normalization_values[j]
    # smoothed_points /= normalization_values

    return smoothed_points


def _masked_smooth_gaussian_sparse(
    sparse_array: np.ndarray,
    positions: np.ndarray,
    sigmas: Union[float, tuple],
    dim_space: int = 3,
):
    """
    Use this function to smooth a sparse array using a Gaussian kernel
    if you expect the output to be a sparse array as well.

    A sparse array is an array of dimension (n_points, n_dim_space + n_dim_points)
    where n_dim_space is the number of spatial dimensions and n_dim_points is the number
    of dimensions of the signal at each point, e.g n_dim_points = 1 for a scalar field,
    n_dim_points = 3 for a 3D vector field, n_dim_points = 9 for a 3x3 matrix field, etc.

    For instance, with a 3D vector field, sparse_array[0] = [z0, y0, x0, v0z, v0y, v0x],
    """

    sigmas = np.array(sigmas).reshape((1, -1))

    old_positions = sparse_array[:, :dim_space]
    positions = positions if positions is not None else old_positions.copy()

    old_tree = KDTree(old_positions / sigmas)
    new_tree = KDTree(positions / sigmas)

    n_dim = sparse_array.shape[1]
    dim_points = n_dim - dim_space
    n_points = positions.shape[0]

    smoothed_array = np.zeros((positions.shape[0], n_dim))
    smoothed_array[:, :dim_space] = positions

    # Find the nearest neighbors of each point
    sparse_dist_matrix = old_tree.sparse_distance_matrix(
        new_tree, max_distance=3, output_type="coo_matrix"
    )
    cols = sparse_dist_matrix.col
    rows = sparse_dist_matrix.row
    dists = sparse_dist_matrix.data
    values = sparse_array[rows, dim_space:]

    smoothed_array[:, dim_space:] = _loop_numba(
        cols, values, dists, n_points, dim_points
    )

    return smoothed_array



from additional_regionprops_properties import (
    add_principal_lengths,
    add_tensor_inertia,
    add_true_strain_tensor,
    add_ellipsoidal_coefficients,
    add_ellipsoidal_nature,
    add_ellipsoidal_nature_bool,
    add_anisotropy_coefficient,
)

from deformation_quantification import tensors_to_napari_vectors

__all__ = [
    "add_principal_lengths",
    "add_tensor_inertia",
    "add_true_strain_tensor",
    "add_ellipsoidal_coefficients",
    "add_ellipsoidal_nature",
    "add_ellipsoidal_nature_bool",
    "add_anisotropy_coefficient",
    "tensors_to_napari_vectors",
]


import numpy as np
from scipy.ndimage import gaussian_filter
from scipy.ndimage.morphology import binary_fill_holes
from skimage.filters import threshold_otsu
from skimage.measure import label
from skimage.morphology import convex_hull_image
from skimage.transform import rescale, resize

from tapenade.preprocessing._smoothing import _masked_smooth_gaussian


def _snp_threshold_binarization(
    image: np.ndarray,
    sigma_blur: float,
    threshold_factor: float,
    registered_image: bool = False,
) -> np.ndarray:
    """
    Threshold image based on signal-and-noise product.

    Parameters:
    - image: numpy array, input image
    - sigma_blur: float, standard deviation of the Gaussian blur.
    - threshold_factor: float, factor to multiply the threshold
    - registered_image: bool, set to True if the image has been registered beforehand
      and thus contains large regions of zeros that lead to sharp intensity gradients
      at the boundaries.

    Returns:
    - binary_mask: numpy array, binary mask computed using SNP thresholding
    """

    if registered_image:
        nonzero_mask = image > 0

        blurred = _masked_smooth_gaussian(image, mask=nonzero_mask, sigmas=sigma_blur)

        blurred2 = _masked_smooth_gaussian(
            image**2, mask=nonzero_mask, sigmas=sigma_blur
        )

        sigma = blurred2 - blurred**2

        snp_array = sigma * blurred
        snp_mask = snp_array > 0

        snp_array = np.log(
            snp_array, where=np.logical_and(nonzero_mask, snp_mask)
        )
    else:
        blurred = _masked_smooth_gaussian(image, sigmas=sigma_blur)
        blurred2 = _masked_smooth_gaussian(image**2, sigmas=sigma_blur)

        sigma = blurred2 - blurred**2

        snp_array = sigma * blurred
        snp_mask = snp_array > 0
        # snp_array = np.log(snp_array, where=(snp_array != 0))

        snp_array = np.log(snp_array, where=snp_mask)

    threshold = threshold_otsu(snp_array[snp_mask]) * threshold_factor

    # Create a binary mask
    binary_mask = np.logical_and(snp_array > threshold, snp_mask)

    return binary_mask


def _otsu_threshold_binarization(
    image: np.ndarray, sigma_blur: float, threshold_factor: float
) -> np.ndarray:
    """
    Threshold image based on histogram values.

    Parameters:
    - image: numpy array, input image
    - sigma_blur: float, standard deviation of the Gaussian blur.
    - threshold_factor: float, factor to multiply the threshold

    Returns:
    - binary_mask: numpy array, binary mask computed using histogram thresholding
    """

    blurred = gaussian_filter(image, sigma=sigma_blur)

    threshold = threshold_otsu(blurred) * threshold_factor

    # Create a binary mask
    binary_mask = blurred > threshold

    return binary_mask


### POST-PROCESSING FUNCTIONS
def _get_largest_connected_component(array: np.ndarray) -> np.ndarray:
    """
    Get the largest connected component in a binary array.

    Parameters:
    - array: Binary array.

    Returns:
    - mask_largest_cc: Binary mask of the largest connected component.
    """
    labels = label(array, connectivity=1)
    mask_largest_cc = labels == np.argmax(np.bincount(labels.flat)[1:]) + 1
    return mask_largest_cc


def _compute_mask_convex_hull(
    mask: np.ndarray, precision_factor: int
) -> np.ndarray:
    """
    Compute the convex hull of the binary mask.

    Parameters:
    - mask: numpy array, binary mask
    - precision_factor: int, factor to rescale the mask before computing the convex hull.
      Bigger values will result in faster computation but less precise convex hulls.

    Returns:
    - hull_mask: numpy array, binary mask representing the convex hull of the input mask
    """

    hull_mask = mask.copy()

    if precision_factor != 1:
        # Rescale the mask if precision_factor is not 1
        hull_mask = rescale(
            hull_mask,
            1 / precision_factor,
            anti_aliasing=False,
            order=0,
            preserve_range=True,
        )

    # Compute the convex hull of the mask
    hull_mask = convex_hull_image(hull_mask)

    if precision_factor != 1:
        # Resize the convex hull mask to the original shape
        hull_mask = resize(
            hull_mask,
            mask.shape,
            anti_aliasing=False,
            order=0,
            preserve_range=True,
        )

    return hull_mask


def _binary_fill_holes_on_each_slice(mask: np.ndarray) -> np.ndarray:
    """
    Fill holes in a binary mask on each slice. Apply the binary_fill_holes function
    to each slice of the input mask twice sequentially.

    Parameters:
    - mask: numpy array, binary mask

    Returns:
    - mask: numpy array, binary mask with holes filled on each slice
    """

    filled_mask = mask.copy()

    for i in range(filled_mask.shape[2]):
        filled_mask[:, :, i] = binary_fill_holes(filled_mask[:, :, i])
    for i in range(filled_mask.shape[1]):
        filled_mask[:, i] = binary_fill_holes(filled_mask[:, i])
    for i in range(mask.shape[0]):
        filled_mask[i] = binary_fill_holes(filled_mask[i])
    # for i in range(filled_mask.shape[0]):
    #     filled_mask[i] = binary_fill_holes(filled_mask[i])
    # for i in range(filled_mask.shape[1]):
    #     filled_mask[:, i] = binary_fill_holes(filled_mask[:, i])
    # for i in range(filled_mask.shape[2]):
    #     filled_mask[:, :, i] = binary_fill_holes(filled_mask[:, :, i])

    return filled_mask


def _refine_raw_mask(
    mask: np.ndarray, post_processing_method: str, keep_largest_cc: bool
) -> np.ndarray:
    """
    Refine the raw mask by keeping only the largest connected component and filling holes.

    Parameters:
    - mask: numpy array, binary mask
    - post_processing_method: str, method to use for post-processing the mask. Can be 'convex_hull' to compute
      the convex hull of the mask, 'fill_holes' to fill holes in each 2D slice of the mask, or 'none' to skip
    - keep_largest_cc: bool, set to True to keep only the largest connected component in the mask.

    Returns:
    - refined_mask: numpy array, refined binary mask
    """
    # Keep only the largest connected component
    if keep_largest_cc:
        refined_mask = _get_largest_connected_component(mask)
    # Fill holes in the mask
    if post_processing_method == "convex_hull":
        refined_mask = _compute_mask_convex_hull(refined_mask, 3)
    elif post_processing_method == "fill_holes":
        refined_mask = binary_fill_holes(refined_mask)
        refined_mask = _binary_fill_holes_on_each_slice(refined_mask)
    else:
        pass

    return refined_mask


###


def _compute_mask(
    image: np.ndarray,
    method: str,
    sigma_blur: float,
    threshold_factor: float = 1,
    # compute_convex_hull: bool = False,
    keep_largest_cc: bool = True,
    post_processing_method: str = "fill_holes",
    registered_image: bool = False,
) -> np.ndarray:
    """
    Process the mask for the given image.

    Parameters:
    - image: numpy array, input image
    - method: str, method to use for thresholding. Can be 'snp otsu' for Signal-Noise Product thresholding,
      or 'otsu' for Otsu's thresholding.
    - sigma_blur: float, standard deviation of the Gaussian blur. Should typically be
      around 1/3 of the typical object diameter.
    - threshold_factor: float, factor to multiply the threshold
    - keep_largest_cc: bool, set to True to keep only the largest connected component in the mask.
    - post_processing_method: str, method to use for post-processing the mask. Can be 'convex_hull' to compute
      the convex hull of the mask, 'fill_holes' to fill holes in each 2D slice of the mask, or 'none' to skip
    - registered_image: bool, set to True if the image has been registered beforehand and thus
      contains large regions of zeros that lead to sharp intensity gradients at the boundaries.

    Returns:
    - mask: numpy array, binary mask of the same shape as the input image
    """

    # Normalize the image
    percs = np.percentile(image, [1, 99])
    im = (image - percs[0]) / (percs[1] - percs[0])
    im = np.clip(im, 0, 1).astype(np.float32)

    # Compute the mask
    if method == "snp otsu":
        mask = _snp_threshold_binarization(
            im / im.max(), sigma_blur, threshold_factor, registered_image
        )
    elif method == "otsu":
        mask = _otsu_threshold_binarization(
            im / im.max(), sigma_blur, threshold_factor
        )
    else:
        raise ValueError(f"Unknown thresholding method: {method}")

    # Refine the mask by keeping only the largest connected component and filling holes
    if keep_largest_cc or post_processing_method != "none":
        mask = _refine_raw_mask(mask, post_processing_method, keep_largest_cc)

    return mask



import numba
import numpy as np


def add_tensor_moment(regionprops_prop, scale):

    regionprops_prop.tensor_moment = prop_tensor_moment(
        regionprops_prop, scale
    )

    return regionprops_prop


def add_tensor_inertia(regionprops_prop, scale):

    regionprops_prop.tensor_inertia = prop_tensor_inertia(
        regionprops_prop, scale
    )

    return regionprops_prop


def add_principal_lengths(
    regionprops_prop, scale, add_principal_vectors: bool = False
):

    if add_principal_vectors:

        principal_lengths, principal_vectors = prop_principal_lengths(
            regionprops_prop, scale=scale, return_principal_vectors=True
        )

        regionprops_prop.principal_lengths = principal_lengths
        regionprops_prop.principal_vectors = principal_vectors

    else:

        principal_lengths = prop_principal_lengths(
            regionprops_prop, scale=scale, return_principal_vectors=False
        )

        regionprops_prop.principal_lengths = principal_lengths

    return regionprops_prop


def add_ellipsoidal_coefficients(regionprops_prop, scale):

    regionprops_prop.ellipsoidal_coefficients = prop_ellipsoidal_coefficients(
        regionprops_prop, scale
    )

    return regionprops_prop


def add_ellipsoidal_nature(
    regionprops_prop, scale, similarity_threshold: float
):

    regionprops_prop.ellipsoidal_nature = prop_ellipsoidal_nature(
        regionprops_prop, scale, similarity_threshold
    )

    return regionprops_prop


def add_ellipsoidal_nature_bool(regionprops_prop, scale):

    regionprops_prop.ellipsoidal_nature_bool = prop_ellipsoidal_nature_bool(
        regionprops_prop, scale
    )

    return regionprops_prop


def add_anisotropy_coefficient(regionprops_prop, scale):

    regionprops_prop.anisotropy_coefficient = prop_anisotropy_coefficient(
        regionprops_prop, scale
    )

    return regionprops_prop


def add_true_strain_tensor(regionprops_prop, scale):

    regionprops_prop.true_strain_tensor = prop_true_strain_tensor(
        regionprops_prop, scale
    )

    return regionprops_prop


def prop_principal_lengths(
    regionprops_prop, scale, return_principal_vectors: bool = False
):
    """
    See https://math.stackexchange.com/questions/2792009/inertia-tensor-of-an-ellipsoid
    """

    if not hasattr(regionprops_prop, "tensor_inertia"):
        regionprops_prop = add_tensor_inertia(regionprops_prop, scale)

    tensor = regionprops_prop.tensor_inertia
    print('ti',tensor)
    rescaled_tensor = tensor * 5 / regionprops_prop.area
    print('tir',rescaled_tensor)

    axis_decoupling_matrix = np.ones((3, 3)) / 2 - np.eye(3)

    if return_principal_vectors:
        eigen_values, principal_vectors = np.linalg.eigh(rescaled_tensor)
        print('ev',eigen_values)
        principal_lengths = np.sqrt(axis_decoupling_matrix @ eigen_values)
        print('pl',principal_lengths)
        return principal_lengths, principal_vectors.T
    else:
        eigen_values = np.linalg.eigvalsh(rescaled_tensor)
        principal_lengths = np.sqrt(axis_decoupling_matrix @ eigen_values)

        return principal_lengths


@numba.jit(nopython=True, nogil=True)
def fast_tensor_inertia(centered_argwheres):
    """
    We apply the Parallel Axis Theorem on the individual voxels:
    each cubic voxel has an individual diagonal inertia tensor
    I = 1/6 * Identity, so its contribution to the inertia tensor
    of the object (on the diagonal) is 1/6 * d^2, where d is the
    distance of the voxel to the center of mass of the object.
    """

    tensor = np.zeros((3, 3), dtype=np.float32)

    for pos in centered_argwheres:

        tensor[0, 0] += pos[1] * pos[1] + pos[2] * pos[2] + 1 / 6  # y^2 + z^2
        tensor[1, 1] += pos[0] * pos[0] + pos[2] * pos[2] + 1 / 6  # x^2 + z^2
        tensor[2, 2] += pos[0] * pos[0] + pos[1] * pos[1] + 1 / 6  # x^2 + y^2

        tensor[0, 1] += -pos[0] * pos[1]
        tensor[0, 2] += -pos[0] * pos[2]
        tensor[1, 0] += -pos[1] * pos[0]
        tensor[1, 2] += -pos[1] * pos[2]
        tensor[2, 0] += -pos[2] * pos[0]
        tensor[2, 1] += -pos[2] * pos[1]

    return tensor


@numba.jit(nopython=True, nogil=True)
def fast_tensor_moment(centered_argwheres):

    tensor = np.zeros((3, 3), dtype=np.float32)

    for pos in centered_argwheres:

        tensor[0, 0] += pos[0] * pos[0]
        tensor[1, 1] += pos[1] * pos[1]
        tensor[2, 2] += pos[2] * pos[2]

        tensor[0, 1] += pos[0] * pos[1]
        tensor[0, 2] += pos[0] * pos[2]
        tensor[1, 0] += pos[1] * pos[0]
        tensor[1, 2] += pos[1] * pos[2]
        tensor[2, 0] += pos[2] * pos[0]
        tensor[2, 1] += pos[2] * pos[1]

    return tensor


def prop_tensor_inertia(regionprops_prop, scale):

    boolean_image = regionprops_prop.image

    argwheres = np.argwhere(boolean_image) * scale
    center = regionprops_prop.centroid_local * scale

    centered_argwheres = argwheres - center

    return fast_tensor_inertia(centered_argwheres)


def prop_tensor_moment(regionprops_prop, scale):

    boolean_image = regionprops_prop.image

    argwheres = np.argwhere(boolean_image) * scale
    center = regionprops_prop.centroid_local * scale

    centered_argwheres = argwheres - center

    return fast_tensor_moment(centered_argwheres)


def prop_true_strain_tensor(regionprops_prop, scale):

    if not hasattr(regionprops_prop, "principal_vectors"):
        regionprops_prop = add_principal_lengths(
            regionprops_prop, scale, add_principal_vectors=True
        )

    principal_lengths = np.array(regionprops_prop.principal_lengths)
    principal_vectors = regionprops_prop.principal_vectors

    denominator = np.power(np.product(principal_lengths), 1 / 3)
    true_strains = np.log(principal_lengths / denominator)

    tensor = (
        principal_vectors.T
        @ np.diag(true_strains)
        @ np.linalg.inv(principal_vectors.T)
    )

    return tensor


def prop_ellipsoidal_nature_bool(regionprops_prop, scale):

    if not hasattr(regionprops_prop, "ellipsoidal_coefficients"):
        regionprops_prop = add_ellipsoidal_coefficients(
            regionprops_prop, scale
        )

    l1, l2, l3 = regionprops_prop.principal_lengths

    R = (l1 * l2 * l3) ** (1 / 3)

    criterion = np.sum((np.array([l1, l2, l3]) - R) > 0)

    return criterion


def prop_ellipsoidal_nature(
    regionprops_prop, scale, similarity_threshold: float
):
    "oblate: earth = 3; prolate: rugby = 2; spheroid = 4; misc = 1"

    if not hasattr(regionprops_prop, "ellipsoidal_coefficients"):
        regionprops_prop = add_ellipsoidal_coefficients(
            regionprops_prop, scale
        )

    alpha, beta, gamma = regionprops_prop.ellipsoidal_coefficients

    alpha = alpha < similarity_threshold
    beta = beta < similarity_threshold
    gamma = gamma < similarity_threshold

    if alpha and beta and gamma:
        return 4
    elif (not alpha) and beta and (not gamma):
        return 3
    elif (not alpha) and (not beta) and gamma:
        return 2
    else:
        return 1


def prop_ellipsoidal_coefficients(regionprops_prop, scale):

    if not hasattr(regionprops_prop, "principal_lengths"):
        regionprops_prop = add_principal_lengths(regionprops_prop, scale)

    l1, l2, l3 = np.sort(regionprops_prop.principal_lengths)

    alpha = (l3 - l1) / (l1 + l3)
    beta = (l3 - l2) / (l2 + l3)
    gamma = (l2 - l1) / (l1 + l2)

    return alpha, beta, gamma


def prop_anisotropy_coefficient(regionprops_prop, scale):
    """
    Ratio of the longest to the shortest principal axis.
    """
    if not hasattr(regionprops_prop, "principal_lengths"):
        regionprops_prop = add_principal_lengths(regionprops_prop, scale)

    l1, _, l3 = regionprops_prop.principal_lengths

    return l1 / l3



from ._preprocessing import (
    change_array_pixelsize,
    compute_mask,
    local_image_equalization,
    normalize_intensity,
    segment_stardist,
    segment_stardist_from_files,
    align_array_major_axis,
    align_array_major_axis_from_files,
    crop_array_using_mask,
    crop_array_using_mask_from_files,
    masked_gaussian_smoothing,
    masked_gaussian_smooth_dense_two_arrays_gpu,
    masked_gaussian_smooth_sparse
)

__all__ = [
    "change_array_pixelsize",
    "compute_mask",
    "local_image_equalization",
    "normalize_intensity",
    "segment_stardist",
    "segment_stardist_from_files",
    "align_array_major_axis",
    "align_array_major_axis_from_files",
    "crop_array_using_mask",
    "crop_array_using_mask_from_files",
    "masked_gaussian_smoothing",
    "masked_gaussian_smooth_dense_two_arrays_gpu",
    "masked_gaussian_smooth_sparse",
]


import concurrent.futures
from functools import partial
from os import cpu_count
from typing import Optional, Union

import numpy as np
import tifffile
from scipy.ndimage import rotate
from skimage.measure import regionprops
from tqdm import tqdm
from tqdm.contrib.concurrent import process_map

from tapenade.preprocessing._array_rescaling import _change_array_pixelsize
from tapenade.preprocessing._axis_alignment import (
    _compute_rotation_angle_and_indices,
)
from tapenade.preprocessing._intensity_normalization import (
    _normalize_intensity,
)
from tapenade.preprocessing._local_equalization import _local_equalization
from tapenade.preprocessing._segmentation import _load_model, _segment_stardist
from tapenade.preprocessing._smoothing import (
    _masked_smooth_gaussian,
    _masked_smooth_gaussian_sparse,
)
from tapenade.preprocessing._thresholding import _compute_mask

"""
#! TODO:
    - refactor every multiprocessing calls through a single decorator
    - update notebook to reflect changes
"""

"""
In typical order:
    1. making array isotropic
    2. compute mask
    3. local image equalization
    (4. segmentation, not covered here)
    (5. spatial registration, not covered here)
    (6. temporal registration, not covered here)
    7. aligning major axis
    (8. cropping array using mask)
"""


# def isotropize_and_normalize(
#     mask, image, labels, scale, sigma: float = None, pos_ref: int = 0
# ):
#     """
#     Make an image isotropic and normalized with respect to a reference channel. Works for multichannel images (ZCYX convention) or single channel images (ZYX convention).
#     Parameters
#     ----------
#     mask : np.array (bool)
#         mask of the image
#     image : np.array
#         image to normalize
#     labels : np.array
#         labels of the mask
#     scale : tuple
#         scale factors for the isotropic transformation
#     sigma : int
#         sigma for the gaussian filter
#     pos_ref : int
#         position of the reference channel, starting from 0
#     Returns
#     -------
#     norm_image : np.array
#         normalized and isotropic image
#     """

#     if len(image.shape) > 3:  # if multichannel image
#         nb_channels = image.shape[1]
#         assert (
#             pos_ref < nb_channels
#         ), "The position of the reference channel is greater than the number of channels. Choose 0 if the first channel is the reference, 1 if the second channel is the reference, etc."
#         iso_image = []
#         liste_channels = np.linspace(
#             0, nb_channels - 1, nb_channels, dtype=int
#         )
#         for ch in liste_channels:
#             channel = image[:, ch, :, :]
#             (mask_iso, channel_iso, seg_iso) = change_arrays_pixelsize(
#                 mask=mask,
#                 image=channel,
#                 labels=labels,
#                 input_pixelsize=scale,
#                 output_pixelsize=(1, 1, 1),
#                 order=1,
#                 n_jobs=-1,
#             )
#             iso_image.append(channel_iso)

#         iso_image = np.array(iso_image)
#         iso_image = iso_image.transpose(1, 0, 2, 3)  # stay in convention ZCYX

#         ref_channel = iso_image[
#             :, pos_ref, :, :
#         ]  # should check before if pos_ref>=iso_image.shape[1]
#         liste_float_channels = np.delete(liste_channels, pos_ref)
#         norm_image = np.zeros_like(iso_image)
#         for ch_float in liste_float_channels:
#             channel = iso_image[:, ch_float, :, :]
#             channel_norm, ref_norm = normalize_intensity(
#                 image=channel,
#                 ref_image=ref_channel,
#                 mask=mask_iso,
#                 labels=seg_iso,
#                 sigma=sigma,
#             )
#             norm_image[:, ch_float, :, :] = channel_norm
#         norm_image[:, pos_ref, :, :] = ref_norm

#     else:  # 3D data, one channel
#         (mask_iso, iso_image, seg_iso) = change_arrays_pixelsize(
#             mask=mask,
#             image=image,
#             labels=labels,
#             reshape_factors=np.divide(scale, (1, 1, 1)),
#         )
#         norm_image, _ = normalize_intensity(
#             image=iso_image,
#             ref_image=iso_image,
#             mask=mask_iso,
#             labels=seg_iso,
#             sigma=sigma,
#         )

#     return (mask_iso, norm_image, seg_iso)


def change_array_pixelsize(
    array: np.ndarray,
    input_pixelsize: tuple[float, float, float] = (1, 1, 1),
    output_pixelsize: tuple[float, float, float] = (1, 1, 1),
    order: int = 1,
    n_jobs: int = -1,
) -> np.ndarray:
    """
    Resizes an input image to have isotropic voxel dimensions.

    Parameters:
    - array: numpy array
    - input_pixelsize: tuple of floats, input pixel dimensions (e.g. in microns)
    - output_pixelsize: tuple of floats, output pixel dimensions (e.g. in microns)
    - order: int, order of interpolation for resizing (defaults to 1 for
      linear interpolation). Choose 0 for nearest-neighbor interpolation
      (e.g. for label images)
    - n_jobs: int, optional number of parallel jobs for resizing (default: -1)

    Returns:
    - resized_array: numpy array
    """

    is_temporal = array.ndim == 4
    n_frames = 0

    if is_temporal:
        if n_jobs == 1:
            # Sequential resizing of each time frame
            resized_array = [
                _change_array_pixelsize(
                    arr, input_pixelsize, output_pixelsize, order=order
                )
                for arr in tqdm(
                    array,
                    desc="Making array isotropic",
                    total=n_frames,
                )
            ]

        else:
            # Parallel resizing of each time frame using multiple processes
            func_parallel = partial(
                _change_array_pixelsize,
                input_pixelsize=input_pixelsize,
                output_pixelsize=output_pixelsize,
                order=order,
            )

            max_workers = (
                cpu_count() if n_jobs == -1 else min(n_jobs, cpu_count())
            )

            resized_array = process_map(
                func_parallel,
                array,
                max_workers=max_workers,
                desc="Changing array pixelsize",
                total=n_frames,
            )

    else:
        # Resizing the whole image
        resized_array = _change_array_pixelsize(
            array,
            input_pixelsize=input_pixelsize,
            output_pixelsize=output_pixelsize,
            order=order,
        )

    return resized_array


def compute_mask(
    image: np.ndarray,
    method: str,
    sigma_blur: float,
    threshold_factor: float = 1,
    post_processing_method: str = "fill_holes",
    keep_largest_cc: bool = True,
    registered_image: bool = False,
    n_jobs: int = -1,
) -> np.ndarray:
    """
    Compute the mask for the given image using the specified method.

    Parameters:
    - image: numpy array, input image
    - method: str, method to use for thresholding. Can be 'snp otsu' for Signal-Noise Product thresholding,
      or 'otsu' for Otsu's thresholding.
    - sigma_blur: float, standard deviation of the Gaussian blur. Should typically be
      around 1/3 of the typical object diameter.
    - threshold_factor: float, factor to multiply the threshold (default: 1)
    - post_processing_method: str, method to use for post-processing the mask. Can be 'convex_hull' to compute
      the convex hull of the mask, 'fill_holes' to fill holes in the mask, or 'none' to skip post-processing
      (default: 'fill_holes')
    - keep_largest_cc: bool, set to True to keep only the largest connected component in the mask (default: True)
    - n_jobs: int, number of parallel jobs to run (-1 for using all available CPUs)

    Returns:
    - mask: numpy array, binary mask of the same shape as the input image
    """

    is_temporal = image.ndim == 4

    if is_temporal:
        func_parallel = partial(
            _compute_mask,
            method=method,
            sigma_blur=sigma_blur,
            threshold_factor=threshold_factor,
            post_processing_method=post_processing_method,
            keep_largest_cc=keep_largest_cc,
            registered_image=registered_image,
        )

        if n_jobs == 1:
            # Sequential processing
            mask = np.array(
                [
                    func_parallel(im)
                    for im in tqdm(image, desc="Thresholding image")
                ]
            )
        else:
            # Parallel processing
            max_workers = (
                cpu_count() if n_jobs == -1 else min(n_jobs, cpu_count())
            )

            mask = np.array(
                process_map(
                    func_parallel,
                    image,
                    max_workers=max_workers,
                    desc="Thresholding image",
                )
            )
    else:
        # Single image processing
        mask = _compute_mask(
            image,
            method=method,
            sigma_blur=sigma_blur,
            threshold_factor=threshold_factor,
            post_processing_method=post_processing_method,
            keep_largest_cc=keep_largest_cc,
            registered_image=registered_image,
        )

    return mask


def local_image_equalization(
    image: np.ndarray,
    box_size: int,
    perc_low: float,
    perc_high: float,
    mask: np.ndarray = None,
    n_jobs: int = -1,
) -> np.ndarray:
    """
    Performs local image equalization on either a single image or a temporal stack of images.
    Stretches the image histogram in local neighborhoods by remapping intesities in the range
    [perc_low, perc_high] to the range [0, 1].
    This helps to enhance the contrast and improve the visibility of structures in the image.

    Parameters:
    - image: numpy array, input image or temporal stack of images
    - box_size: int, size of the local neighborhood for equalization
    - perc_low: float, lower percentile for intensity equalization
    - perc_high: float, upper percentile for intensity equalization
    - mask: numpy array, binary mask used to set the background to zero (optional)
    - n_jobs: int, number of parallel jobs to use (not used currently as the function is parallelized internally)

    Returns:
    - image_norm: numpy array, equalized image or stack of equalized images
    """

    is_temporal = image.ndim == 4

    mask_is_None = mask is None

    if is_temporal:
        if mask_is_None:
            mask = [None] * image.shape[0]

        # Apply local equalization to each time frame in the temporal stack
        image_norm = np.array(
            [
                _local_equalization(
                    image[ind_t],
                    box_size=box_size,
                    perc_low=perc_low,
                    perc_high=perc_high,
                    mask=mask[ind_t],
                )
                for ind_t in tqdm(
                    range(image.shape[0]), desc="Local equalization"
                )
            ]
        )
    else:
        # Apply local equalization to the image
        image_norm = _local_equalization(
            image,
            box_size=box_size,
            perc_low=perc_low,
            perc_high=perc_high,
            mask=mask,
        )

    if not mask_is_None:
        # Set the background to zero using the mask
        image_norm = np.where(mask, image_norm, 0.0)

    return np.clip(image_norm, 0, 1)


def normalize_intensity(
    image: np.ndarray,
    ref_image: np.ndarray,
    sigma: float = None,
    mask: np.ndarray = None,
    labels: np.ndarray = None,
    width: int = 3,
    n_jobs: int = -1,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Normalize the intensity of an image based on a reference image assumed to have
    ideally homogeneous signal (e.g DAPI).

    Parameters:
    - image: numpy array, input image to be normalized
    - ref_image: numpy array, reference image used for normalization
    - sigma: float, standard deviation for Gaussian smoothing of the reference
        image (default: None)
    - mask: numpy array, binary mask of the sample (default: None)
    - labels: numpy array, labels indicating the instances in which the reference
        signal is expressed (default: None)
    - width: int, number of neighboring planes to consider for reference plane
        calculation (default: 3)
    - n_jobs: int, number of parallel jobs to use (-1 for all available CPUs, 1 for
        sequential execution) (default: -1)

    Returns:
    - image_norm: numpy array, normalized input image
    - ref_image_norm: numpy array, normalized reference image
    """

    is_temporal = image.ndim == 4

    if is_temporal:
        if mask is None:
            mask = [None] * image.shape[0]

        if labels is None:
            labels = [None] * image.shape[0]

        if n_jobs == 1:
            # Sequential normalization of each time frame
            normalized_array = [
                _normalize_intensity(
                    im,
                    ref_im,
                    sigma=sigma,
                    mask=ma,
                    labels=lab,
                    width=width,
                )
                for im, ref_im, ma, lab in tqdm(
                    zip(image, ref_image, mask, labels, strict=False),
                    desc="Normalizing intensity",
                    total=image.shape[0],
                )
            ]

        else:
            # Parallel normalization of each time frame using multiple processes
            func_parallel = partial(
                _normalize_intensity,
                sigma=sigma,
                mask=mask,
                labels=labels,
                width=width,
            )

            max_workers = (
                cpu_count() if n_jobs == -1 else min(n_jobs, cpu_count())
            )

            normalized_array = process_map(
                func_parallel,
                zip(image, ref_image, strict=False),
                max_workers=max_workers,
                desc="Normalizing intensity",
                total=image.shape[0],
            )

    else:
        # Single image normalization
        normalized_array = _normalize_intensity(
            image,
            ref_image,
            sigma=sigma,
            mask=mask,
            labels=labels,
            width=width,
        )

    return np.array(normalized_array)


def segment_stardist(
    image: np.ndarray,
    model_path: str,
    thresholds_dict: dict = None,
    n_jobs: int = -1,
) -> np.ndarray:
    """
    Predict the segmentation of an array using a StarDist model.

    Parameters:
    - image: a 3D numpy array, input image to segment
    - model_path: str, path to the StarDist model folder
    - thresholds_dict: dict, dictionary of thresholds for the model, structured like
      {'prob': 0.5, 'nms': 0.3} for probability and non-maximum suppression thresholds
      respectively (default: None)
    - n_jobs: int, not used here but kept for consistency with other functions
    """

    is_temporal = image.ndim == 4

    model = _load_model(model_path)

    if is_temporal:
        labels = np.array(
            [
                _segment_stardist(
                    im,
                    model=model,
                    thresholds_dict=thresholds_dict,
                )
                for im in tqdm(image, desc="Predicting with StarDist")
            ],
            dtype=np.uint16,
        )

    else:
        labels = _segment_stardist(
            image,
            model=model,
            thresholds_dict=thresholds_dict,
        )

    return labels


def segment_stardist_from_files(
    image_files: list[str],
    path_to_save: str,
    compress_params: dict,
    func_params: dict,
):

    model = _load_model(func_params["model_path"])

    for file in image_files:
        image = tifffile.imread(file)
        labels = _segment_stardist(
            image,
            model=model,
            thresholds_dict=func_params.get("thresholds_dict"),
        )
        tifffile.imwrite(
            f"{path_to_save}/segmented_{file}", labels, **compress_params
        )


def align_array_major_axis(
    target_axis: str,
    rotation_plane: str,
    mask: np.ndarray,
    array: np.ndarray,
    order: int = 1,
    temporal_slice: Optional[int] = None,
    n_jobs: int = -1,
) -> np.ndarray:
    """
    Aligns the major axis of an array to a target axis in a specified rotation plane.
    This function uses Principal Component Analysis (PCA) to determine the major axis of the array,
    and then rotates the array to align the major axis with the target axis.

    Parameters:
    - target_axis: str, the target axis to align the major axis with ('X', 'Y', or 'Z')
    - rotation_plane: str, the rotation plane to perform the rotation in ('XY', 'XZ', or 'YZ')
    - array: numpy array
    - order: int, order of interpolation for image rotation (default: 1)
    - temporal_slice: int, optional temporal slicing applied to the mask before computing its major axis (default: None)
    - n_jobs: int, number of parallel jobs to use (-1 for all available CPUs, 1 for sequential execution) (default: -1)

    Returns:
    - rotated_array: numpy array, array with the major axis aligned to the target axis
    """

    is_temporal = mask.ndim == 4

    temporal_slice = slice(None) if temporal_slice is None else temporal_slice

    # Compute the mask that will be used to compute the major axis.
    # If the mask is a temporal array, the major axis is computed by aggregating the mask
    # instances in temporal_slice.
    mask_for_pca = (
        np.any(mask[temporal_slice], axis=0) if is_temporal else mask
    )

    # Compute the rotation angle and the indices of the rotation plane
    rotation_angle, rotation_plane_indices = (
        _compute_rotation_angle_and_indices(
            mask_for_pca, target_axis, rotation_plane
        )
    )

    if array.dtype == bool:
        order = 0

    # Define the rotation functions
    func_rotate = partial(
        rotate,
        angle=rotation_angle,
        axes=rotation_plane_indices,
        reshape=True,
        order=order,
    )

    func_rotate_mask = partial(
        rotate,
        angle=rotation_angle,
        axes=rotation_plane_indices,
        reshape=True,
        order=0,
    )

    if is_temporal and n_jobs != 1:
        # Rotate the arrays in parallel if the array is temporal and parallel execution is enabled
        max_workers = cpu_count() if n_jobs == -1 else min(n_jobs, cpu_count())

        array_rotated = np.array(
            process_map(
                func_rotate,
                array,
                max_workers=max_workers,
                desc="Aligning array",
            )
        )

        mask_rotated = np.array(
            process_map(
                func_rotate_mask,
                mask,
                max_workers=max_workers,
                desc="Aligning mask",
            )
        )

    else:
        # Rotate the arrays in block
        array_rotated = func_rotate(array)
        mask_rotated = func_rotate_mask(mask)

    array_rotated = np.where(mask_rotated, array_rotated, 0)

    return array_rotated


def _load_array_rotate_and_save_to_file(
    array_file: str,
    mask_file: str,
    index: int,
    path_to_save: str,
    rotation_angle: float,
    order: int,
    rotation_plane_indices: tuple[int, int],
    compress_params: dict,
):
    array = tifffile.imread(array_file)
    mask = tifffile.imread(mask_file)
    array_rotated = rotate(
        array,
        angle=rotation_angle,
        axes=rotation_plane_indices,
        reshape=True,
        order=order,
    )
    mask_rotated = rotate(
        mask,
        angle=rotation_angle,
        axes=rotation_plane_indices,
        reshape=True,
        order=0,
    )
    array_rotated = np.where(mask_rotated, array_rotated, 0)
    if order > 1:
        # preserve the original intensity range
        array_rotated = np.clip(array_rotated, array.min(), array.max())

    tifffile.imwrite(
        f"{path_to_save}/aligned_{index:>04}.tif",
        array_rotated,
        **compress_params,
    )


def align_array_major_axis_from_files(
    mask_files: list[str],
    array_files: list[str],
    path_to_save: str,
    compress_params: dict,
    func_params: dict,
):
    """
    Aligns the major axis of an array to a target axis in a specified rotation plane.
    This function uses Principal Component Analysis (PCA) to determine the major axis of the array,
    and then rotates the array to align the major axis with the target axis.
    """

    mask_zyx_shape = tifffile.imread(mask_files[0]).shape
    mask_for_pca = np.zeros(mask_zyx_shape[:-3], dtype=bool)

    for mask_file in mask_files:
        mask = tifffile.imread(mask_file)
        mask_for_pca = np.logical_or(mask_for_pca, mask)

    target_axis = func_params.get("target_axis", "Z")
    rotation_plane = func_params.get("rotation_plane", "XY")
    order = func_params.get("order", 1)

    # Compute the rotation angle and the indices of the rotation plane
    rotation_angle, rotation_plane_indices = (
        _compute_rotation_angle_and_indices(
            mask_for_pca, target_axis, rotation_plane
        )
    )

    multithreaded_function = partial(
        _load_array_rotate_and_save_to_file,
        path_to_save=path_to_save,
        order=order,
        rotation_angle=rotation_angle,
        rotation_plane_indices=rotation_plane_indices,
        compress_params=compress_params,
    )

    n_jobs = func_params.get("n_jobs", -1)

    # open all array files using the multithreading library and crop the results
    with concurrent.futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
        tqdm(
            executor.map(
                multithreaded_function,
                array_files,
                mask_files,
                range(len(array_files)),
            ),
            total=len(array_files),
            desc="Aligning array",
        )


def crop_array_using_mask(
    mask: np.ndarray,
    array: np.ndarray,
    margin: int = 0,
    n_jobs: int = -1,
) -> np.ndarray:
    """
    Crop an array using a binary mask. If the array is temporal, the cropping
    slice is computed by aggregating mask instances at all times.

    Parameters:
    - mask: numpy array, binary mask indicating the region of interest
    - array: numpy array, array to crop based on the mask
    - margin: int, optional margin to add around the mask (default: 0)
    - n_jobs: int, number of parallel jobs to use (not used currently as the function is not computationally intensive)

    Returns:
    - cropped_array: numpy array, cropped array based on the mask
    """

    is_temporal = mask.ndim == 4

    # Compute aggregated mask if array is temporal
    mask_for_slice = np.any(mask, axis=0) if is_temporal else mask

    # Get the mask slice
    mask_slice = regionprops(mask_for_slice.astype(int))[0].slice

    mask_zyx_shape = mask.shape[1:] if is_temporal else mask.shape

    # Add margin to the slice if specified
    if margin > 0:
        mask_slice = tuple(
            slice(
                max(0, mask_slice[i].start - margin),
                min(mask_slice[i].stop + margin, mask_zyx_shape[i]),
            )
            for i in range(3)
        )

    mask_slice = (slice(None),) + mask_slice if is_temporal else mask_slice

    # Apply the slice to the array
    array_cropped = array[mask_slice]

    return array_cropped


def _extract_slice_from_file(file: str):
    """
    Extract the bounding box from a file containing a binary mask.
    """

    mask = tifffile.imread(file)
    mask_slice = regionprops(mask.astype(int))[0].slice

    return mask_slice


def _load_array_crop_and_save_to_file(
    array_file: str,
    index: int,
    path_to_save: str,
    mask_slice: tuple[slice],
    compress_params: dict,
):
    array = tifffile.imread(array_file)[mask_slice]
    tifffile.imwrite(
        f"{path_to_save}/cropped_{index:>04}.tif", array, **compress_params
    )


def crop_array_using_mask_from_files(
    mask_files: list[str],
    array_files: list[str],
    path_to_save: str,
    compress_params: dict,
    func_params: dict,
) -> np.ndarray:
    """
    Crop an array using a binary mask. If the array is temporal, the cropping
    slice is computed by aggregating mask instances at all times.
    """

    mask_zyx_shape = tifffile.imread(mask_files[0]).shape

    # open all mask files using the multithreading library
    mask_slices = []

    n_jobs = func_params.get("n_jobs", -1)

    with concurrent.futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
        mask_slices = list(
            tqdm(
                executor.map(_extract_slice_from_file, mask_files),
                total=len(mask_files),
                desc="Extracting mask slices",
            )
        )

    # aggregate the mask slices
    mask_slice = tuple(
        slice(
            min([mask_slice[i].start for mask_slice in mask_slices]),
            max([mask_slice[i].stop for mask_slice in mask_slices]),
        )
        for i in range(3)
    )

    margin = func_params.get("margin", 0)

    # Add margin to the slice if specified
    if margin > 0:
        mask_slice = tuple(
            slice(
                max(0, mask_slice[i].start - margin),
                min(mask_slice[i].stop + margin, mask_zyx_shape[i]),
            )
            for i in range(3)
        )

    multithreaded_function = partial(
        _load_array_crop_and_save_to_file,
        path_to_save=path_to_save,
        mask_slice=mask_slice,
        compress_params=compress_params,
    )
    # open all array files using the multithreading library and crop the results
    with concurrent.futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
        tqdm(
            executor.map(
                multithreaded_function,
                array_files,
                range(len(array_files)),
            ),
            total=len(array_files),
            desc="Cropping array",
        )


def _parallel_gaussian_smooth(
    input_tuple: tuple[np.ndarray, np.ndarray],
    sigmas: Union[float, list[float]],
) -> np.ndarray:
    data, mask, mask_for_volume = input_tuple
    return _masked_smooth_gaussian(data, sigmas, mask, mask_for_volume)


def masked_gaussian_smoothing(
    image: np.ndarray,
    sigmas: Union[float, list[float]],
    mask: Optional[np.ndarray] = None,
    mask_for_volume: Optional[np.ndarray] = None,
    n_jobs: int = -1,
) -> np.ndarray:
    """
    Apply Gaussian smoothing to an image or a sequence of images.

    Parameters:
        image (ndarray): The input image or sequence of images.
        sigmas (float or list of floats): The standard deviation(s) of the Gaussian kernel.
        mask (ndarray, optional): The mask indicating the regions of interest. Default is None.
        mask_for_volume (ndarray, optional): The mask indicating the regions of interest for volume calculation. Default is None.
        n_jobs (int, optional): The number of parallel jobs to run. Default is -1, which uses all available CPU cores.

    Returns:
        ndarray: The smoothed image or sequence of images.
    """

    is_temporal = image.ndim == 4

    if is_temporal:

        if mask is None:
            mask = [None] * image.shape[0]

        if mask_for_volume is None:
            mask_for_volume = [None] * image.shape[0]

        func = partial(_parallel_gaussian_smooth, sigmas=sigmas)

        if n_jobs == 1:

            iterable = tqdm(
                zip(image, mask, mask_for_volume, strict=False),
                total=len(image),
                desc="Smoothing image",
            )

            return np.array([func(elem) for elem in iterable])

        else:
            elems = list(zip(image, mask, strict=False))

            max_workers = (
                cpu_count() if n_jobs == -1 else min(n_jobs, cpu_count())
            )
            result = process_map(
                func, elems, max_workers=max_workers, desc="Smoothing image"
            )

            return np.array(result)

    else:
        return _masked_smooth_gaussian(image, sigmas, mask, mask_for_volume)


def masked_gaussian_smooth_dense_two_arrays_gpu(
    datas: list[np.ndarray],
    sigmas: Union[float, list[float]],
    mask: np.ndarray = None,
    masks_for_volume: Union[np.ndarray, list[np.ndarray]] = None,
):
    """
    Inputs in data are assumed to be non-temporal, 3D arrays.
    """

    from pyclesperanto_prototype import gaussian_blur

    if isinstance(sigmas, int | float):
        sigmas = [sigmas] * 3

    if mask is None:
        smoothed1 = np.array(
            gaussian_blur(
                datas[0].astype(np.float16),
                sigma_x=sigmas[0],
                sigma_y=sigmas[1],
                sigma_z=sigmas[2],
            )
        )
        smoothed2 = np.array(
            gaussian_blur(
                datas[1].astype(np.float16),
                sigma_x=sigmas[0],
                sigma_y=sigmas[1],
                sigma_z=sigmas[2],
            )
        )

    elif masks_for_volume is None:

        smoothed1 = np.array(
            gaussian_blur(
                np.where(mask, datas[0].astype(np.float16), 0.0),
                sigma_x=sigmas[0],
                sigma_y=sigmas[1],
                sigma_z=sigmas[2],
            )
        )

        smoothed2 = np.array(
            gaussian_blur(
                np.where(mask, datas[1].astype(np.float16), 0.0),
                sigma_x=sigmas[0],
                sigma_y=sigmas[1],
                sigma_z=sigmas[2],
            )
        )

        effective_volume = np.array(
            gaussian_blur(
                mask.astype(np.float16),
                sigma_x=sigmas[0],
                sigma_y=sigmas[1],
                sigma_z=sigmas[2],
            )
        )

        smoothed1 = np.where(
            mask, np.divide(smoothed1, effective_volume, where=mask), 0.0
        )

        smoothed2 = np.where(
            mask, np.divide(smoothed2, effective_volume, where=mask), 0.0
        )

    else:  # both masks and masks_for_volume are not None

        if isinstance(masks_for_volume, np.ndarray):

            smoothed1 = np.array(
                gaussian_blur(
                    np.where(
                        masks_for_volume, datas[0].astype(np.float16), 0.0
                    ),
                    sigma_x=sigmas[0],
                    sigma_y=sigmas[1],
                    sigma_z=sigmas[2],
                )
            )

            smoothed2 = np.array(
                gaussian_blur(
                    np.where(
                        masks_for_volume, datas[1].astype(np.float16), 0.0
                    ),
                    sigma_x=sigmas[0],
                    sigma_y=sigmas[1],
                    sigma_z=sigmas[2],
                )
            )

            effective_volume = np.array(
                gaussian_blur(
                    masks_for_volume.astype(np.float16),
                    sigma_x=sigmas[0],
                    sigma_y=sigmas[1],
                    sigma_z=sigmas[2],
                )
            )

            smoothed1 = np.where(
                mask, np.divide(smoothed1, effective_volume, where=mask), 0.0
            )

            smoothed2 = np.where(
                mask, np.divide(smoothed2, effective_volume, where=mask), 0.0
            )

        elif isinstance(masks_for_volume, list):

            smoothed1 = np.array(
                gaussian_blur(
                    np.where(
                        masks_for_volume[0], datas[0].astype(np.float16), 0.0
                    ),
                    sigma_x=sigmas[0],
                    sigma_y=sigmas[1],
                    sigma_z=sigmas[2],
                )
            )

            smoothed2 = np.array(
                gaussian_blur(
                    np.where(
                        masks_for_volume[1], datas[1].astype(np.float16), 0.0
                    ),
                    sigma_x=sigmas[0],
                    sigma_y=sigmas[1],
                    sigma_z=sigmas[2],
                )
            )

            effective_volume1 = np.array(
                gaussian_blur(
                    masks_for_volume[0].astype(np.float16),
                    sigma_x=sigmas[0],
                    sigma_y=sigmas[1],
                    sigma_z=sigmas[2],
                )
            )

            effective_volume2 = np.array(
                gaussian_blur(
                    masks_for_volume[1].astype(np.float16),
                    sigma_x=sigmas[0],
                    sigma_y=sigmas[1],
                    sigma_z=sigmas[2],
                )
            )

            smoothed1 = np.where(
                mask, np.divide(smoothed1, effective_volume1, where=mask), 0.0
            )

            smoothed2 = np.where(
                mask, np.divide(smoothed2, effective_volume2, where=mask), 0.0
            )

    return smoothed1, smoothed2


def masked_gaussian_smooth_sparse(
    sparse_array: Union[np.ndarray, list[np.ndarray]],
    is_temporal: bool,
    dim_space: int,
    sigmas: Union[float, tuple[float]],
    positions: np.ndarray = None,
    n_job: int = -1,
    progress_bars: bool = True,
):
    """
    Smooth sparse data using a gaussian kernel.
    Sparse data is a numpy array with the first columns being the spatial coordinates,
    and the last columns being values of interest.

    Parameters
    ----------
    sparse_array : np.ndarray or list of np.ndarray
        Array of points to smooth in format (n_points, n_dim_space + n_dim_points).
        The first columns (up to dim_space, i.e [:dim_space]) must be the spatial coordinates.
        The remaining columns are the values/vectors to smooth. A temporal sparse array is a
        list of sparse arrays, one for each time point.
    is_temporal : bool
        If True, the array is temporal and the smoothing is applied to each time step.
    dim_space : int
        Number of spatial dimensions.
    sigmas : float or list of float
        Standard deviations of the gaussian kernel.
    positions : np.ndarray, optional
        Positions where the smoothing is applied. If None, the smoothing is applied to the
        positions of the input array.
    mask : np.ndarray, optional
        Mask to apply to the positions. If None, no mask is applied.
    n_job : int, optional
        Number of jobs to run in parallel. If -1, all the available CPUs are used.
        The default is -1.

    Returns
    -------
    np.ndarray
        The smoothed sparse array.
    """

    positions_is_temporal = isinstance(positions, list)

    if is_temporal:

        if n_job == 1:

            if positions_is_temporal:
                return np.array(
                    [
                        _masked_smooth_gaussian_sparse(
                            elem, pos, sigmas, dim_space
                        )
                        for elem, pos in tqdm(
                            zip(
                                sparse_array,
                                positions,
                                disable=not progress_bars, strict=False,
                            )
                        )
                    ]
                )

            else:
                return np.array(
                    [
                        _masked_smooth_gaussian_sparse(
                            elem, positions, sigmas, dim_space
                        )
                        for elem in tqdm(
                            sparse_array, disable=not progress_bars
                        )
                    ]
                )

        else:
            if not positions_is_temporal:
                # same positions for all time steps
                func = partial(
                    _masked_smooth_gaussian_sparse,
                    sigmas=sigmas,
                    dim_space=dim_space,
                    positions=positions,
                )

                max_workers = (
                    cpu_count() if n_job == -1 else min(n_job, cpu_count())
                )
                result = np.array(
                    process_map(
                        func,
                        sparse_array,
                        max_workers=max_workers,
                        disable=not progress_bars,
                    )
                )

            else:
                func = partial(
                    _masked_smooth_gaussian_sparse,
                    sigmas=sigmas,
                    dim_space=dim_space,
                )

                max_workers = (
                    cpu_count() if n_job == -1 else min(n_job, cpu_count())
                )
                result = process_map(
                    func,
                    sparse_array,
                    positions,
                    max_workers=max_workers,
                    disable=not progress_bars,
                )

            return result

    else:
        return _masked_smooth_gaussian_sparse(
            array=sparse_array,
            positions=positions,
            sigmas=sigmas,
            dim_space=dim_space,
        )



from os import cpu_count

import numpy as np
from tqdm.contrib.concurrent import process_map

from tapenade.preprocessing._labels_masking import (
    _remove_labels_outside_of_mask,
)


def remove_labels_outside_of_mask(
    mask: np.ndarray, labels: np.ndarray, n_jobs: int = -1
) -> np.ndarray:
    """
    Removes labels outside (or at the border) of the mask.

    Parameters:
    - mask (ndarray): The mask indicating the valid region.
    - labels (ndarray): The segmentation labels.
    - n_jobs (int): The number of parallel jobs to run. If -1, use all available CPUs.

    Returns:
        ndarray: The post-processed segmentation labels.
    """
    is_temporal = labels.ndim == 4

    if is_temporal:

        if n_jobs == 1:
            # Process each label and mask pair sequentially
            labels_filtered = np.array(
                [
                    _remove_labels_outside_of_mask(lab, ma)
                    for lab, ma in zip(labels, mask, strict=False)
                ]
            )
        else:
            max_workers = (
                cpu_count() if n_jobs == -1 else min(n_jobs, cpu_count())
            )

            # Process each label and mask pair in parallel using multiple workers
            labels_filtered = np.array(
                process_map(
                    _remove_labels_outside_of_mask,
                    labels,
                    mask,
                    max_workers=max_workers,
                    desc="Removing labels outside of mask",
                )
            )

    else:
        # Process the single label and mask pair
        labels_filtered = _remove_labels_outside_of_mask(labels, mask)

    return labels_filtered



from itertools import product

import numpy as np
from scipy.ndimage import zoom


def _local_equalization(
    image: np.ndarray,
    box_size: int,
    perc_low: float,
    perc_high: float,
    mask: np.ndarray = None,
) -> np.ndarray:
    """
    Performs local histogram stretching by applying the following steps:
        1. compute the percentile values in a neighborhood around voxels chosen on a regular grid.
        2. interpolate these values on each voxel of the image.
        3. perform a linear equalization between the interpolated percentile values and the values 0 and 1.
        4. clip the image voxel values between 0 and 1.

    Parameters
    ----------
    image : np.ndarray
        The image to equalize.
    box_size : int
        size of the neighborhood box around each voxel to compute the percentile values.
    perc_low : float
        percentile value to use as the low value of the equalization (will be mapped to 0).
    perc_high : float
        percentile value to use as the high value of the equalization (will be mapped to 1).
    mask : np.ndarray
        An optional mask boolean array from which background values will be excluded from the computation.

    Returns
    -------
    image_norm : np.ndarray
        The equalized image.
    """

    # Compute necessary variables
    array_shape = image.shape
    box_length = 2 * box_size + 1
    grid_shape = [int(np.ceil(s / box_length)) + 1 for s in array_shape]
    grid_positions = [
        np.linspace(0, s - 1, n_boxes)
        for s, n_boxes in zip(array_shape, grid_shape, strict=False)
    ]

    percs_low = np.zeros(grid_shape)
    percs_high = np.zeros(grid_shape)

    percs_positions = [np.arange(s) for s in grid_shape]

    # Compute percentile values for each neighborhood
    for indices_grid, indices_percs in zip(
        product(*grid_positions), product(*percs_positions), strict=False
    ):
        indices_grid = np.array(indices_grid).round().astype(int)

        slices = []
        for index, s in zip(indices_grid, array_shape, strict=False):
            start = min(max(index - box_size, 0), s - box_length)
            stop = min(max(index + box_size + 1, box_length), s)
            slices.append(slice(start, stop))

        box = image[tuple(slices)].copy()

        # for debug, actually should not be necessary
        if box.size < 0.2 * box_length**image.ndim:
            continue

        if mask is not None:
            box = box[mask[tuple(slices)]]
            if box.size == 0:
                continue

        val_low, val_high = np.percentile(box, [perc_low, perc_high])

        percs_low[tuple(indices_percs)] = val_low
        percs_high[tuple(indices_percs)] = val_high

    # Interpolate percentile values for each voxel
    full_percs_low = zoom(
        percs_low,
        zoom=[s / n for s, n in zip(array_shape, grid_shape, strict=False)],
        order=1,
    )

    full_percs_high = zoom(
        percs_high,
        zoom=[s / n for s, n in zip(array_shape, grid_shape, strict=False)],
        order=1,
    )

    denom = full_percs_high - full_percs_low
    denom[denom == 0] = 1
    denom += 1e-8

    # Perform linear equalization and clip image voxel values
    # image_norm = np.clip((image - full_percs_low) / denom, 0, 1)
    image_norm = (image - full_percs_low) / denom

    return image_norm



import numpy as np
from skimage.measure import regionprops


def _remove_labels_outside_of_mask(
    labels: np.ndarray, mask: np.ndarray
) -> np.ndarray:
    """
    Helper function to remove labels outside (or at the border) of the mask.

    Args:
        labels (ndarray): The segmentation labels.
        mask (ndarray): The mask indicating the valid region.

    Returns:
        ndarray: The post-processed segmentation labels.
    """
    props = regionprops(labels)

    for prop in props:
        mask_roi = mask[prop.slice]

        # Check if the label is at least partially outside the mask
        if np.any(~mask_roi):

            # Calculate the volume of the label that intersects the mask
            volume_inside = np.logical_and(prop.image, mask_roi).sum()

            # If the volume inside is smaller than the total volume of the label,
            # remove the label from the labels array
            if volume_inside < prop.area:
                labels_roi = labels[prop.slice]
                labels_roi[labels_roi == prop.index] = 0

    return labels



import numpy as np
from skimage.transform import resize
from tqdm import tqdm

from tapenade.preprocessing import change_arrays_pixelsize, segment_stardist


def predict_stardist(
    array: np.ndarray,
    model_path: str,
    input_voxelsize: tuple,
    voxelsize_model: tuple = (0.7, 0.7, 0.7),
    normalize_input: bool = True,
):
    """
    Predict the segmentation of an array using a StarDist model.

    Parameters:
    - array: a 3D numpy array, input image to segment
    - model_path: str, path to the StarDist model
    - input_voxelsize: tuple, input voxel size of the image
    - voxelsize_model: tuple, voxel size of the model
    - normalize_input: bool, whether to normalize the input image

    Returns:
    - aniso_labels: numpy array, predicted
    """
    assert len(np.shape(array)) <= 3

    data = change_arrays_pixelsize(
        image=array,
        input_pixelsize=input_voxelsize,
        output_pixelsize=voxelsize_model,
        order=1,
    )

    if normalize_input:
        perc_low, perc_high = np.percentile(data, (1, 99))
        data = (data - perc_low) / (perc_high - perc_low)

    labels = segment_stardist(data, model_path)

    aniso_labels = resize(
        labels,
        array.shape,
        anti_aliasing=False,
        order=0,
        preserve_range=True,
    )

    return aniso_labels.astype(np.int16)


def find_seg_errors(segmentation: np.ndarray, image: np.ndarray):
    """
    Compute statisitics of intensity of each label to try to find out exceptions/anomalies and detect segmentation errors.

    Parameters:
    segmentation: array containing the labels
    image : intensity image, has to be the same size as segmentatio,

    Returns:
    The intensity_distribution array, for which each line corresponds to a label in the segmentation.
    Column 1 is the label's id
    Col 2 is its mean intensity (in the ROI)
    Col 3 is the std of its intensity distrib
    Col 4 is the ratio of standard deviation and mean. We use this value to detect wrong segmentations

    """
    list_labels = list(np.unique(segmentation))
    list_labels.remove(0)
    print(len(list_labels), " labels to process")

    intensity_distribution = np.zeros((len(list_labels), 4))
    for index, label in tqdm(enumerate(list_labels)):
        intensity_distribution[index, 0] = label
        mask = segmentation == label
        list_pix = image[mask]
        n, bins = np.histogram(list_pix)
        mids = 0.5 * (bins[1:] + bins[:-1])
        mean = np.average(mids, weights=n)
        var = np.average((mids - mean) ** 2, weights=n)
        intensity_distribution[index, 1] = mean
        intensity_distribution[index, 2] = var
        intensity_distribution[index, 3] = var / mean

    return intensity_distribution


def tresh_distribution(
    intensity_distribution: np.ndarray,
    threshold: float,
    column_number: int = 3,
):
    """
    Thresholds the Distribution to find labels that could be wrong segmentation.

    Parameters:
    intensity_distribution: Array computed with the function find_seg_errors, that shows average intensity, std and std/mean for each label of the distrbution
    threshold : threshold value above which cell will be considered wrong. To get an idea, visualize the distrbution (examples are in the notebooks)
    column_number : index to consider for thre threshold.
                    If you want to discrimnate on the mean intensity of the ROI, choose 1.
                    If you want to discriminate on the std, choose 2.
                    If you want to discriminate over the std divided by the mean intensity for each label, choose 3.

    Returns:
    A list of labels that might be false segmentations.

    """

    id_merged_cells = []
    for index, intensity in enumerate(
        intensity_distribution[:, column_number]
    ):
        if intensity > threshold:
            id_merged_cells.append(int(intensity_distribution[index, 0]))
    return id_merged_cells



import numpy as np
from scipy.optimize import minimize_scalar

from tapenade.preprocessing._smoothing import _masked_smooth_gaussian


def _median_absolute_deviation(array: np.ndarray) -> float:
    return np.nanmedian(np.abs(array - np.nanmedian(array)))


def _nans_outside_mask(array: np.ndarray, mask: np.ndarray):
    return np.where(mask, array, np.nan)


def _optimize_sigma(
    ref_array: np.ndarray, mask: np.ndarray, labels_mask: np.ndarray
):

    def opt_func(
        sigma, ref_array: np.ndarray, mask: np.ndarray, labels_mask: np.ndarray
    ):
        ref_array_smooth = _masked_smooth_gaussian(
            ref_array, sigmas=sigma, mask_for_volume=labels_mask, mask=mask
        )

        if labels_mask is not None:
            res = np.divide(ref_array, ref_array_smooth, where=labels_mask)
            res = _nans_outside_mask(res, labels_mask)
        elif mask is not None:
            res = np.divide(ref_array, ref_array_smooth, where=mask)
            res = _nans_outside_mask(res, mask)
        else:
            res = ref_array / ref_array_smooth

        return _median_absolute_deviation(np.nanmedian(res, axis=(1, 2)))

    res = minimize_scalar(
        opt_func,
        args=(ref_array, mask, labels_mask),
        bounds=(10, 30),
        method="bounded",
        options={"maxiter": 5, "disp": 3},
    )

    return res.x


def _find_reference_plane_from_medians(array: np.ndarray):
    ref_ind = np.nanargmax(np.nanmedian(array, axis=(1, 2)))
    return ref_ind


def _normalize_intensity(
    array: np.ndarray,
    ref_array: np.ndarray,
    sigma=None,
    mask: np.ndarray = None,
    labels: np.ndarray = None,
    width=3,
):
    """
    Normalize the intensity of an array based on a reference array assumed to have
    ideally homogeneous signal (e.g DAPI).

    Parameters:
    - array (ndarray): The input array to be normalized.
    - ref_array (ndarray): The reference array used for normalization.
    - sigma (float, optional): The standard deviation for Gaussian smoothing of the reference array. Default is None,
        but a value is required right now since setting 'sigma' to None is not implemented.
    - mask (ndarray, optional): A binary mask of the sample. Default is None.
    - labels (ndarray, optional): An array of labels indicating the instances in which the reference
        signal is expressed, e.g nuclei. Default is None.
    - width (int, optional): The number of neighboring planes to consider for reference plane calculation. Default is 5.

    Returns:
    - array_norm (ndarray): The normalized input array.
    - ref_array_norm (ndarray): The normalized reference array.
    """

    num_z_slices = array.shape[0]

    labels_mask = None if labels is None else labels.astype(bool)

    # compute smoothed reference array for normalization
    if sigma is None:
        sigma = _optimize_sigma(ref_array, mask, labels_mask)
        print("sigma = ", sigma)
    ref_array_smooth = _masked_smooth_gaussian(
        ref_array, sigmas=sigma, mask_for_volume=labels_mask, mask=mask
    )

    if mask is not None:
        array = _nans_outside_mask(array, mask)
        ref_array = _nans_outside_mask(ref_array, mask)
        ref_array_smooth = _nans_outside_mask(ref_array_smooth, mask)

        mask_divide = mask
    else:
        mask_divide = True

    # normalize array and reference array
    array_norm = np.divide(array, ref_array_smooth, where=mask_divide)
    ref_array_norm = np.divide(ref_array, ref_array_smooth, where=mask_divide)

    if mask is not None:
        array_norm = _nans_outside_mask(array_norm, mask)
        ref_array_norm = _nans_outside_mask(ref_array_norm, mask)

    # rectify median intensity in both normalized arrays
    # to that of the median of the brightest consecutive planes
    z_ref = _find_reference_plane_from_medians(ref_array)
    z_ref_norm = _find_reference_plane_from_medians(ref_array_norm)

    sl = slice(max(0, z_ref - width), min(num_z_slices, z_ref + width))
    sl_norm = slice(
        max(0, z_ref_norm - width), min(num_z_slices, z_ref_norm + width)
    )

    array_normalization_factor = np.nanmedian(
        array_norm[sl_norm]
    ) / np.nanmedian(array[sl])

    array_norm = array_norm / array_normalization_factor

    return array_norm



import warnings

import numpy as np

try:
    from stardist import gputools_available
    from stardist.models import StarDist3D
except ImportError:
    warnings.warn("Please install Stardist for your system")

import os
from pathlib import Path


def _load_model(model_path: str):
    """
    Load a StarDist model from a folder.

    Parameters:
    - model_path: str, path to the StarDist model folder
    """

    model_name = Path(model_path).stem
    directory = str(os.path.split(model_path)[0])
    model = StarDist3D(None, name=model_name, basedir=directory)

    return model


def _segment_stardist(image: np.ndarray, model, thresholds_dict: dict):
    """
    Predict the segmentation of an array using a StarDist model.

    Parameters:
    - image: a 3D numpy array, input image to segment
    - model: StarDist model object
    - model_path: str, path to the StarDist model folder
    - thresholds_dict: dict, dictionary of thresholds for the model, structured like
      {'prob': 0.5, 'nms': 0.3} for probability and non-maximum suppression thresholds
      respectively
    """

    if gputools_available():
        model.use_gpu = True

    if thresholds_dict is not None:
        model.thresholds = thresholds_dict

    labels, _ = model.predict_instances(
        image, n_tiles=model._guess_n_tiles(image)
    )

    return labels.astype(np.uint16)



from ._reconstruct import (
    add_centermass,
    associate_positions,
    check_napari,
    create_folders,
    extract_positions,
    fuse_sides,
    transformation_from_plugin,
    manual_registration_fct,
    plot_positions,
    register,
    remove_previous_files,
    sigmoid,
    write_hyperstacks,
)

__all__ = (
    "extract_positions",
    "plot_positions",
    "associate_positions",
    "transformation_from_plugin"
    "manual_registration_fct",
    "register",
    "create_folders",
    "check_napari",
    "fuse_sides",
    "sigmoid",
    "write_hyperstacks",
    "add_centermass",
    "remove_previous_files",
)
